# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import json
import logging
import math
import os
import tempfile
from concurrent.futures import ProcessPoolExecutor
from functools import partial
from glob import glob
from typing import Iterable, Optional, cast

import numpy as np
from composer.utils import (
    ObjectStore,
    maybe_create_object_store_from_uri,
    parse_uri,
)
from numpy.typing import NDArray
from streaming import MDSWriter
from tqdm import tqdm
from transformers import AutoTokenizer, PreTrainedTokenizerBase

from llmfoundry.data.data import AbstractConcatTokensDataset
from llmfoundry.utils.data_prep_utils import (
    DownloadingIterable,
    download_file,
    merge_shard_groups,
)
from llmfoundry.utils.exceptions import (
    CannotUnicodeDecodeFile,
    DatasetTooSmallError,
    InputFolderMissingDataError,
    InputFolderNotFound,
    OutputFolderNotEmptyError,
)

log = logging.getLogger(__name__)

DONE_FILENAME = '.text_to_mds_conversion_done'


class ConcatTokensFromFilesDataset(AbstractConcatTokensDataset):
    """An IterableDataset that returns token samples for MDSWriter from files.

    Returns dicts of {'tokens': ndarray:int32}

    Each file is considered a sequence.
    """

    def __init__(
        self,
        files: Iterable[str],
        tokenizer: PreTrainedTokenizerBase,
        max_length: int,
        bos_text: str,
        eos_text: str,
        no_wrap: bool,
    ):
        self.files = files
        super().__init__(tokenizer, max_length, bos_text, eos_text, no_wrap)
        log.info(f'Initialized ConcatTokensFromFilesDataset.')

    def __iter__(self) -> Iterable[dict[str, NDArray]]:
        log.info(
            'Starting iteration over files in ConcatTokensFromFilesDataset',
        )
        buffer = []
        for file in self.files:
            log.info(f'Processing file: {file}')
            with open(file, 'r') as f:
                buffer += self.bos_tokens
                first_chunk = True
                # Read the file in 1MB chunks to avoid memory issues
                try:
                    for chunk in iter(partial(f.read, 1000000), ''):
                        # Tokenize the chunk
                        encoded = self.tokenizer(
                            chunk,
                            truncation=False,
                            padding=False,
                        )
                        iids = encoded['input_ids']

                        # If this is not the first chunk, remove the BOS token
                        if not first_chunk:
                            if iids[0] == self.tokenizer.bos_token_id:
                                iids = iids[1:]

                        # Add the tokens to the buffer
                        buffer += iids
                        while len(buffer) >= self.max_length:
                            concat_sample = buffer[:self.max_length]
                            buffer = buffer[self.max_length:
                                           ] if self.should_wrap else []
                            yield {
                                'tokens':
                                    np.asarray(concat_sample, dtype=np.int32),
                            }

                        first_chunk = False
                except UnicodeDecodeError:
                    raise CannotUnicodeDecodeFile(text_file=file)

                # Add the EOS token to the buffer to separate files.
                buffer += self.eos_tokens

        # Yield any remaining samples of size max_length.
        while len(buffer) >= self.max_length:
            concat_sample = buffer[:self.max_length]
            buffer = buffer[self.max_length:] if self.should_wrap else []
            yield {'tokens': np.asarray(concat_sample, dtype=np.int32)}

        log.info(
            'Finished iterating over files in ConcatTokensFromFilesDataset',
        )


def get_object_names(input_folder: str) -> list[str]:
    """Get object names from a local or remote folder.

    Args:
        input_folder (str): local or remote folder path.
    """
    object_store = maybe_create_object_store_from_uri(input_folder)
    if object_store is not None:
        _, _, folder_prefix = parse_uri(input_folder)
        try:
            names = [
                name for name in object_store.list_objects(folder_prefix)
                if name.endswith('.txt')
            ]
            log.info(f'Found {len(names)} text files in remote storage')
        except FileNotFoundError:
            raise InputFolderNotFound(folder_prefix)

    else:
        # input_folder is a local folder
        names = [
            text_file for dirpath, _, _ in os.walk(input_folder)
            for text_file in glob(os.path.join(dirpath, '*.txt'))
        ]
    # return names, sizes
    log.info(f'Found {len(names)} text files at {input_folder}')

    return names


def get_task_args(
    object_names: list[str],
    output_root: str,
    input_folder: str,
    n_groups: int,
    tokenizer_name: str,
    concat_tokens: int,
    eos_text: str,
    bos_text: str,
    no_wrap: bool,
    compression: str,
    trust_remote_code: bool,
) -> Iterable:
    """Get download_and_convert arguments split across n_groups.

    Each group handles a portion of object_names.

    Args:
        object_names (List[str]): Names of objects to process
        output_root (str): Folder to write MDS shards to
        input_folder (str): Folder of text files to process
        n_groups (int): Number of groups to split the object names into
        tokenizer_name (str): Name of tokenizer to use
        concat_tokens (int): Concatenate up to this many tokens
        eos_text (str): Text to append to each example to separate concatenated samples
        bos_text (str): Text to prepend to each example to separate concatenated samples
        no_wrap: (bool): Whether to let text examples wrap across multiple training examples
        compression (str): The compression algorithm to use for MDS writing
        trust_remote_code (bool): If true, allows custom code to be executed to load the tokenizer
    """
    log.info(
        f'Preparing task arguments for {len(object_names)} objects across {n_groups} groups',
    )
    num_objects = len(object_names)
    objs_per_group = math.ceil(num_objects / n_groups)
    for group, i in enumerate(range(0, num_objects, objs_per_group)):
        output_subdir = os.path.join(output_root, str(group))
        log.info(
            f'Created task for group {group} with {min(objs_per_group, num_objects - i)} objects',
        )
        yield (
            object_names[i:min(i + objs_per_group, num_objects)],
            output_subdir,
            input_folder,
            tokenizer_name,
            concat_tokens,
            eos_text,
            bos_text,
            no_wrap,
            compression,
            trust_remote_code,
        )


def download_and_convert_starargs(args: tuple):
    """Helper function to call download_and_convert with star args.

    This helps us use download_and_convert with multiprocessing.
    """
    return download_and_convert(*args)


def download_and_convert(
    file_names: list[str],
    output_folder: str,
    input_folder: str,
    tokenizer_name: str,
    concat_tokens: int,
    eos_text: str,
    bos_text: str,
    no_wrap: bool,
    compression: str,
    trust_remote_code: bool,
):
    """Downloads and converts text files to MDS format.

    Args:
        file_names (List[str]): Files to process
        output_folder (str): Folder to write MDS shards to
        input_folder (str): Folder of text files to process
        tokenizer_name (str): Name of tokenizer to use
        concat_tokens (int): Concatenate up to this many tokens
        eos_text (str): Text to append to each example to separate concatenated samples
        bos_text (str): Text to prepend to each example to separate concatenated samples
        no_wrap: (bool): Whether to let text examples wrap across multiple training examples
        compression (str): The compression algorithm to use for MDS writing
        trust_remote_code (bool): If true, allows custom code to be executed to load the tokenizer
    """
    log.info(f'Starting download and conversion for {len(file_names)} files')

    object_store = maybe_create_object_store_from_uri(input_folder)

    # Download file_names
    downloading_iter = DownloadingIterable(
        object_names=file_names,
        output_folder=None, # Downloads to temporary files.
        object_store=object_store,
    )
    log.info(f'Initializing tokenizer: {tokenizer_name}')
    tokenizer = AutoTokenizer.from_pretrained(
        tokenizer_name,
        trust_remote_code=trust_remote_code,
    )
    tokenizer.model_max_length = 5000000000  # Hack to prevent warnings from HuggingFace

    # Use the ConcatTokensDataset from LLM-foundry to concatenate sequences of tokens up
    # to the maximum sequence length
    dataset = ConcatTokensFromFilesDataset(
        files=downloading_iter,
        max_length=concat_tokens,
        tokenizer=tokenizer,
        eos_text=eos_text,
        bos_text=bos_text,
        no_wrap=no_wrap,
    )

    columns = {'tokens': 'ndarray:int32'}

    log.info('Converting to MDS format...')
    with MDSWriter(
        out=output_folder,
        columns=columns,
        compression=compression,
    ) as out:
        for sample in tqdm(dataset):
            out.write(sample)

    log.info(f'Completed download and conversion for {len(file_names)} files')


def is_remote_path(path: str) -> bool:
    """Checks whether a path is a remote path.

    Args:
        path (str): path to check
    """
    backend, _, _ = parse_uri(path)
    return backend != ''


def is_already_processed(
    output_root: str,
    args_str: str,
    object_names: list[str],
) -> bool:
    """Determines whether a group of text files has already been processed.

    Checks the done fie at output root to determine this.

    Args:
        output_root (str): Output folder where a done file may exist
        args_str (str): String representation of the arguments
        object_names (List[str]): Names of objects to convert to MDS format
    """
    log.info(
        f'Checking if {len(object_names)} objects have already been processed in {output_root}',
    )

    # Retrieve the done file contents
    output_object_store = maybe_create_object_store_from_uri(output_root)
    if output_object_store is not None:
        # Download and read the done file from the remote object store
        _, _, output_folder_prefix = parse_uri(output_root)
        try:
            with tempfile.TemporaryDirectory() as tmp_dir:
                done_file = os.path.join(tmp_dir, DONE_FILENAME)
                download_file(
                    object_store=output_object_store,
                    object_name=os.path.join(
                        output_folder_prefix,
                        DONE_FILENAME,
                    ),
                    output_filename=done_file,
                )
                with open(done_file) as df:
                    done_file_contents = df.read().splitlines()
                log.info(f'Retrieved done file contents from remote storage')
        except FileNotFoundError:
            log.info('Done file not found in remote storage')
            return False
    else:
        # Read the local done file
        done_file = os.path.join(output_root, DONE_FILENAME)
        if not os.path.isfile(done_file):
            log.info('Done file not found in local storage')
            return False
        with open(done_file) as df:
            done_file_contents = df.read().splitlines()
        log.info(f'Retrieved done file contents from local storage')

    # Compare the arguments
    prev_args_str = done_file_contents[0]
    if prev_args_str != args_str:
        log.info('Arguments have changed, reprocessing required')
        return False

    # Compare file names
    prev_names = done_file_contents[1:]
    if len(prev_names) != len(object_names):
        log.info('Number of files has changed, reprocessing required')
        return False
    for idx, prev_name in enumerate(prev_names):
        if object_names[idx] != prev_name:
            log.info('File names have changed, reprocessing required')
            return False

    log.info('All files have already been processed')
    return True


def write_done_file(folder: str, args_str: str, object_names: list[str]):
    """Write a file to signify completion.

    This the done file includes the arguments to processing and
    a list of objects that were processed.

    Args:
        folder (str): Folder to write the done file to
        args_str (str): String representation of arguments
        object_names (List[str]): List of objects to convert to MDS format
    """
    with open(os.path.join(folder, DONE_FILENAME), 'w') as done_file:
        log.info(f'Writing done file.')
        done_file.write('\n'.join([args_str] + object_names) + '\n')
    log.info(f'Done file written successfully')


def convert_text_to_mds(
    tokenizer_name: str,
    output_folder: str,
    input_folder: str,
    concat_tokens: int,
    eos_text: str,
    bos_text: str,
    no_wrap: bool,
    compression: str,
    processes: int,
    args_str: str,
    reprocess: bool,
    trust_remote_code: bool,
):
    """Convert a folder of text files to MDS format.

    Args:
        tokenizer_name (str): Name of tokenizer to use
        output_folder (str): Folder to write MDS shards to
        input_folder (str): Folder of text files to process
        concat_tokens (int): Concatenate up to this many tokens
        eos_text (str): Text to append to each example to separate concatenated samples
        bos_text (str): Text to prepend to each example to separate concatenated samples
        no_wrap: (bool): Whether to let text examples wrap across multiple training examples
        compression (str): The compression algorithm to use for MDS writing
        processes (int): The number of processes to use.
        args_str (str): String representation of the arguments
        reprocess (bool): Whether to always reprocess the given folder of text files
        trust_remote_code (bool): If true, allows custom code to be executed to load the tokenizer
    """
    # Load the tokenizer once on the main process so that the files are cached to avoid race conditions
    # in the Hugging Face load code
    AutoTokenizer.from_pretrained(
        tokenizer_name,
        trust_remote_code=trust_remote_code,
    )

    is_remote_output = is_remote_path(output_folder)
    log.info(f'Output is remote: {is_remote_output}')

    object_names = get_object_names(input_folder)
    if len(object_names) == 0:
        log.error(f'No text files found in input folder: {input_folder}')
        raise InputFolderMissingDataError(input_folder)

    # Check if the text files in the bucket have already been processed.
    if not reprocess and is_already_processed(
        output_folder,
        args_str,
        object_names,
    ):
        log.info(
            f'Input folder {input_folder} is already processed at {output_folder} and '
            +
            'reprocess is set to False. Set reprocess to True if you would like to force reprocessing.',
        )
        return

    # Use a temporary local directory if the output is remote and there are more than 1 processes
    local_output_folder = tempfile.TemporaryDirectory(
    ).name if is_remote_output else output_folder
    log.info(f'Using local output folder: {local_output_folder}')

    if os.path.isdir(output_folder) and len(os.listdir(output_folder)) > 0:
        log.error(f'Output folder is not empty: {output_folder}')
        raise OutputFolderNotEmptyError(output_folder)

    if processes > 1:
        log.info(f'Using multiprocessing with {processes} processes')
        # Download and convert the text files in parallel
        args = get_task_args(
            object_names,
            local_output_folder,
            input_folder,
            processes,
            tokenizer_name,
            concat_tokens,
            eos_text,
            bos_text,
            no_wrap,
            compression,
            trust_remote_code,
        )
        with ProcessPoolExecutor(max_workers=processes) as executor:
            list(executor.map(download_and_convert_starargs, args))

        log.info('Merging MDS shards from each process')
        # Merge the mds shards from each of the processes into a single folder
        merge_shard_groups(local_output_folder)
    else:
        log.info('Using single process for download and conversion')
        download_and_convert(
            object_names,
            local_output_folder,
            input_folder,
            tokenizer_name,
            concat_tokens,
            eos_text,
            bos_text,
            no_wrap,
            compression,
            trust_remote_code,
        )

    index_path = os.path.join(local_output_folder, 'index.json')
    with open(index_path, 'r') as index_file:
        if not json.load(index_file)['shards']:
            raise DatasetTooSmallError(
                reason='No shards were created when converting text to MDS.',
            )

    # Write a done file with the args and object names
    write_done_file(local_output_folder, args_str, object_names)

    if is_remote_output:
        # Upload the local output to the remote location
        output_object_store = cast(
            ObjectStore,
            maybe_create_object_store_from_uri(output_folder),
        )
        _, _, output_folder_prefix = parse_uri(output_folder)
        files_to_upload = os.listdir(local_output_folder)

        for file in files_to_upload:
            assert not os.path.isdir(file)
            remote_path = os.path.join(output_folder_prefix, file)
            output_object_store.upload_object(
                remote_path,
                os.path.join(local_output_folder, file),
            )


def _configure_logging(logging_level: str):
    """Configure logging.

    Args:
        logging_level (str): Logging level.
    """
    logging.basicConfig(
        format=
        f'%(asctime)s: [%(process)d][%(threadName)s]: %(levelname)s: %(name)s: %(message)s',
        force=True,
    )
    logging_level = logging_level.upper()
    logging.getLogger('llmfoundry').setLevel(logging_level)
    logging.getLogger(__name__).setLevel(logging_level)
    log.info(f'Logging level set to {logging_level}')


def convert_text_to_mds_from_args(
    output_folder: str,
    input_folder: str,
    compression: str,
    concat_tokens: int,
    tokenizer_name: str,
    bos_text: Optional[str],
    eos_text: Optional[str],
    use_tokenizer_eos: bool,
    no_wrap: bool,
    processes: int,
    reprocess: bool,
    trust_remote_code: bool,
    logging_level: str,
) -> None:
    """A wrapper for `convert_text_to_mds` to parse arguments.

    Args:
        output_folder (str): Folder to write MDS shards to
        input_folder (str): Folder of text files to process
        compression (str): The compression algorithm to use for MDS writing
        concat_tokens (int): Concatenate up to this many tokens
        tokenizer_name (str): The name of the tokenizer to use
        bos_text (Optional[str]): The text to prepend to each example to separate concatenated examples
        eos_text (Optional[str]): The text to append to each example to separate concatenated examples
        use_tokenizer_eos (bool): Use the EOS text from the tokenizer
        no_wrap (bool): Whether to let text examples wrap across multiple training examples
        processes (int): The number of processes to use to download and convert the dataset
        reprocess (bool): If true, reprocess the input_folder to MDS format. Otherwise, only reprocess upon changes to the input folder or dataset creation parameters.
        trust_remote_code (bool): If true, allows custom code to be executed to load the tokenizer
        logging_level (str): Logging level for the script. Default is INFO.

    Raises:
        ValueError: If `use_tokenizer_eos` is True and `eos_text` is not None
    """
    os.environ['WORLD_SIZE'] = '1'
    if use_tokenizer_eos:
        # Ensure that eos text is not specified twice.
        if eos_text is not None:
            ValueError(
                'Cannot set --eos_text with --use_tokenizer_eos. Please specify one.',
            )
        tokenizer = AutoTokenizer.from_pretrained(
            tokenizer_name,
            trust_remote_code=trust_remote_code,
        )
        eos_text = tokenizer.eos_token

    # now that we have validated them, change BOS/EOS to strings
    if bos_text is None:
        bos_text = ''
    if eos_text is None:
        eos_text = ''
    _configure_logging(logging_level)

    # Define args for _args_str
    args = {
        'tokenizer': tokenizer_name,
        'output_folder': output_folder,
        'input_folder': input_folder,
        'compression': compression,
        'concat_tokens': concat_tokens,
        'eos_text': eos_text,
        'bos_text': bos_text,
        'no_wrap': no_wrap,
        'processes': processes,
        'reprocess': reprocess,
        'trust_remote_code': trust_remote_code,
    }
    convert_text_to_mds(
        tokenizer_name=tokenizer_name,
        output_folder=output_folder,
        input_folder=input_folder,
        concat_tokens=concat_tokens,
        eos_text=eos_text,
        bos_text=bos_text,
        no_wrap=no_wrap,
        compression=compression,
        processes=processes,
        reprocess=reprocess,
        trust_remote_code=trust_remote_code,
        args_str=str(args),
    )



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import json
import os
import platform
import warnings
from typing import Any, Callable, Iterable, Optional, Union

import datasets as hf_datasets
import psutil
from datasets import Dataset, DatasetDict, IterableDataset, IterableDatasetDict
from streaming import MDSWriter
from torch.utils.data import DataLoader
from tqdm import tqdm

from llmfoundry.data.finetuning.collator import validate_target_settings
from llmfoundry.data.finetuning.tasks import (
    _get_example_type,
    dataset_constructor,
    is_valid_ift_example,
    tokenize_formatted_example,
)
from llmfoundry.utils.builders import build_tokenizer

HFDataset = Union[DatasetDict, Dataset, IterableDatasetDict, IterableDataset]


def build_dataloader(
    dataset: HFDataset,
    batch_size: int,
    num_workers: Optional[int] = None,
) -> DataLoader:
    if num_workers is None:
        # Multiple workers is only supported on linux machines
        if 'linux' in platform.platform().lower():
            num_workers = max(1, psutil.cpu_count())
        else:
            num_workers = 0

    # If using multiple workers, configure each worker to prefetch as many samples as it can, up to
    # the aggregate device batch size
    # If not using workers, the torch DataLoader expects the default value for prefetch_factor,
    # which non-intuitively must be 2.
    # If on macOS, PyTorch requires prefetch_factor set to None since num_workers is always zero
    if 'macos' in platform.platform().lower() and num_workers == 0:
        prefetch_factor = None
    else:
        prefetch_factor = max(
            1,
            2 * batch_size // num_workers,
        ) if num_workers > 0 else 2

    return DataLoader(
        dataset=dataset,
        sampler=None,
        batch_size=batch_size,
        num_workers=num_workers,
        prefetch_factor=prefetch_factor,
    )


def generate_samples(
    loader: DataLoader,
    truncate_num_samples: Optional[int] = None,
) -> Iterable[dict[str, bytes]]:
    """Generator over samples of a dataloader.

    Args:
       loader (DataLoader): A dataloader emitting batches like {key: [sample0_bytes, sample1_bytes, sample2_bytes, ...]}
       truncate_num_samples (Optional[int]): An optional # of samples to stop at.

    Yields:
        Sample dicts.
    """
    n_samples = 0
    for batch in loader:
        keys = list(batch.keys())
        current_bs = len(batch[keys[0]])
        for idx in range(current_bs):
            if truncate_num_samples is not None and n_samples == truncate_num_samples:
                return
            n_samples += 1
            yield {k: v[idx] for k, v in batch.items()}


def get_columns_and_format(
    dataset: HFDataset,
    tokenizing: bool,
    preprocessing_fn: Callable,
):
    ex = preprocessing_fn(next(iter(dataset)))
    example_type = _get_example_type(ex)
    if tokenizing:
        return {'turns': 'json'}, example_type
    if example_type == 'chat':
        # Chat format
        return {'messages': 'json'}, example_type
    else:
        # Prompt-response format
        return {'prompt': 'str', 'response': 'str'}, example_type


def convert_finetuning_dataset(
    dataset: str,
    data_subset: Optional[str],
    splits: list[str],
    preprocessor: Optional[str],
    data_files: list[str],
    skip_preprocessing: bool,
    out_root: str,
    local: Optional[str],
    compression: Optional[str],
    num_workers: Optional[int],
    tokenizer: Optional[str],
    tokenizer_kwargs: dict[str, Any],
    max_seq_len: int,
    target_prompts: str,
    target_responses: str,
    encoder_decoder: bool,
) -> None:
    """Converts Finetuning datasets to MDS format.

    Args:
        dataset (str): Name of the dataset (e.g., first argument to `datasets.load_dataset`, for jsonl data format, it is `json`).
        data_subset (Optional[str]): Subset of data to use.
        splits (list[str]): Comma-separated list of dataset splits
        preprocessor (Optional[str]): Name or import path of function used to preprocess (reformat) the dataset.
        data_files (list[str]): Data file for each split. Comma-separated.
        skip_preprocessing (bool): Whether to skip preprocessing.
        out_root (str): Root path of output directory where MDS shards will be stored. Can be a remote URI.
        local (Optional[str]): Root path of local directory if you want to keep a local copy when out_root is remote.
        compression (Optional[str]): Name of compression algorithm to use.
        num_workers (Optional[int]): Number of workers.
        tokenizer (Optional[str]): Tokenizer used for processing.
        tokenizer_kwargs (dict[str, Any]): Keyword arguments for tokenizer initialization.
        max_seq_len (int): Maximum sequence length.
        target_prompts (str): Policy for when to use prompts as training targets.
        target_responses (str): Policy for which responses to treat as training targets.
        encoder_decoder (bool): Set if the data are intended to be used to train an encoder-decoder model

    Raises:
        ValueError: If the target settings are invalid.
    """
    if skip_preprocessing:
        preprocessing_fn = lambda x: x  # Just an identity function
    else:
        preprocessor_str = preprocessor
        preprocessing_fn = dataset_constructor.get_preprocessing_fn_from_str(
            preprocessor=preprocessor_str,
            dataset_name=dataset,
        )
        if preprocessing_fn is None:
            raise ValueError(
                '`preprocessor` was not set and no preprocessing function ' +\
                'has been registered for `dataset`. If this was intentional ' +\
                '(e.g., because your dataset is already correctly formatted), ' +\
                'include the "--skip-preprocessing" flag to avoid this error.',
            )

    # Make sure the target settings are valid
    validate_target_settings(
        target_prompts=target_prompts,
        target_responses=target_responses,
        decoder_only_format=not encoder_decoder,
    )

    tokenizer_kwargs = tokenizer_kwargs
    tokenizer_kwargs.update({'model_max_length': max_seq_len})
    if tokenizer:
        tokenizer = build_tokenizer(tokenizer, tokenizer_kwargs)

    for i, split_name in enumerate(splits):
        data_file = None
        if len(data_files) > 0:
            data_file = data_files[i]
        loaded_dataset = hf_datasets.load_dataset(
            path=dataset,
            name=data_subset,
            split=split_name,
            data_files=data_file,
            streaming=True,
        )
        # Determine the output columns
        columns, example_type = get_columns_and_format(
            dataset=loaded_dataset,
            tokenizing=tokenizer is not None,
            preprocessing_fn=preprocessing_fn,
        )
        # Prepare the iterables
        if example_type == 'chat':
            samples = iter(loaded_dataset)
        else:
            loader = build_dataloader(
                dataset=loaded_dataset,
                batch_size=512,
                num_workers=num_workers,
            )
            samples = generate_samples(loader)

        # Write samples
        print(f'Converting {split_name} to MDS format...')
        out = os.path.join(out_root, split_name)
        if local is not None:
            out = (os.path.join(local, split_name), out)
            keep_local = True
        else:
            keep_local = False
        with MDSWriter(
            columns=columns,
            out=out,
            compression=compression,
            keep_local=keep_local,
        ) as out:
            examples_removed = 0
            for sample in tqdm(samples, desc=split_name):
                formatted_sample = preprocessing_fn(sample)
                assert isinstance(formatted_sample, dict)

                # Use the _get_example_type utility to confirm that the formatted sample
                # can be interpreted by the tokenization code
                try:
                    example_type = _get_example_type(formatted_sample)
                except Exception as e:
                    raise ValueError(
                        'Encountered an error when checking example for proper formatting. ' +\
                        f'example={formatted_sample}',
                    ) from e
                if tokenizer is not None:
                    sample = tokenize_formatted_example(
                        formatted_sample,
                        tokenizer=tokenizer,
                    )
                    if not is_valid_ift_example(
                        max_seq_len,
                        target_prompts=target_prompts,
                        target_responses=target_responses,
                        decoder_only_format=not encoder_decoder,
                        example=sample,
                    ):
                        examples_removed += 1
                        continue

                    sample_to_write = {'turns': []}
                    for turn in sample['turns']:
                        turn_to_write = {}
                        for key in ['input_ids', 'labels']:
                            turn_to_write[key] = list(turn[key])
                        sample_to_write['turns'].append(turn_to_write)
                    out.write(sample_to_write)
                else:
                    if example_type == 'prompt_response':
                        encoded_sample = {}
                        for key in ['prompt', 'response']:
                            value = formatted_sample[key]
                            assert isinstance(value, str)
                            encoded_sample[key] = value.encode('utf-8')
                        out.write(encoded_sample)
                    else:
                        out.write(formatted_sample)

        if tokenizer is not None and examples_removed > 0:
            warnings.warn(
                f'Dropped {examples_removed} examples where the prompt was longer than {max_seq_len}, '
                +
                'the prompt or response was empty, or the response was all padding tokens.',
            )


def convert_finetuning_dataset_from_args(
    dataset: str,
    data_subset: Optional[str],
    splits: list[str],
    preprocessor: Optional[str],
    data_files: list[str],
    skip_preprocessing: bool,
    out_root: str,
    local: Optional[str],
    compression: Optional[str],
    num_workers: Optional[int],
    tokenizer: Optional[str],
    tokenizer_kwargs: Optional[str],
    max_seq_len: int,
    target_prompts: str,
    target_responses: str,
    encoder_decoder: bool,
):
    """A wrapper for `convert_finetuning_dataset` to parse arguments.

    Args:
        dataset (str): Name of the dataset (e.g., first argument to `datasets.load_dataset`, for jsonl data format, it is `json`).
        data_subset (Optional[str]): Subset of data to use.
        splits (list[str]): Comma-separated list of dataset splits
        preprocessor (Optional[str]): Name or import path of function used to preprocess (reformat) the dataset.
        data_files (list[str]): Data file for each split. Comma-separated.
        skip_preprocessing (bool): Whether to skip preprocessing.
        out_root (str): Root path of output directory where MDS shards will be stored. Can be a remote URI.
        local (Optional[str]): Root path of local directory if you want to keep a local copy when out_root is remote.
        compression (Optional[str]): Name of compression algorithm to use.
        num_workers (Optional[int]): Number of workers.
        tokenizer (Optional[str]): Tokenizer used for processing.
        tokenizer_kwargs (Optional[str]): Keyword arguments for tokenizer initialization in JSON format.
        max_seq_len (int): Maximum sequence length.
        target_prompts (str): Policy for when to use prompts as training targets.
        target_responses (str): Policy for which responses to treat as training targets.
        encoder_decoder (bool): Set if the data are intended to be used to train an encoder-decoder model.

    Raises:
        ValueError: If the target settings are invalid.
        ValueError: If the output directory already contains the requested splits.
    """
    os.environ['WORLD_SIZE'] = '1'
    if os.path.isdir(out_root) and len(
        set(os.listdir(out_root)).intersection(set(splits)),
    ) > 0:
        raise ValueError(
            f'--out_root={out_root} contains {os.listdir(out_root)} which cannot overlap with the requested splits {splits}.',
        )

    if tokenizer_kwargs is not None:
        parsed_tokenizer_kwargs = json.loads(tokenizer_kwargs)
    else:
        parsed_tokenizer_kwargs = {}

    if len(data_files) > 0 and len(data_files,) != len(splits):
        raise ValueError(
            f'If data_files is set, data_files and splits must have the same length. Got {len(data_files)=} while {len(splits)=}',
        )
    convert_finetuning_dataset(
        dataset=dataset,
        data_subset=data_subset,
        splits=splits,
        preprocessor=preprocessor,
        data_files=data_files,
        skip_preprocessing=skip_preprocessing,
        out_root=out_root,
        local=local,
        compression=compression,
        num_workers=num_workers,
        tokenizer=tokenizer,
        tokenizer_kwargs=parsed_tokenizer_kwargs,
        max_seq_len=max_seq_len,
        target_prompts=target_prompts,
        target_responses=target_responses,
        encoder_decoder=encoder_decoder,
    )



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from typing import Any, Callable, Iterable, Optional

import torch


class NoOp(torch.optim.Optimizer):
    """Optimizer that performs no optimization steps."""

    def __init__(
        self,
        params: Iterable[torch.Tensor],
    ):
        """Initialize NoOp optimizer.

        Args:
            params (Iterable[torch.Tensor]): Model parameters for the optimizer.
        """
        # LR schedulers expect param groups to have LR. Unused.
        defaults = {'lr': 0.0}
        super().__init__(params, defaults)

    def __setstate__(self, state: dict[str, dict[Any, Any]]) -> None:
        super().__setstate__(state)

    def state_dict(self):
        return super().state_dict()

    @torch.no_grad()
    def step(self, closure: Optional[Callable] = None):
        """Perform no-op optimization step where no parameters are updated.

        Args:
            closure (Callable, optional): A closure that reevaluates the model
                and returns the loss.
        """
        loss = None
        if closure is not None:
            with torch.enable_grad():
                loss = closure()

        return loss



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import abc
from typing import Any

from composer.core import Callback

__all__ = ['CallbackWithConfig']


class CallbackWithConfig(Callback, abc.ABC):
    """A callback that takes a config dictionary as an argument, in addition to.

    its other kwargs.
    """

    def __init__(
        self,
        config: dict[str, Any],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        del config, args, kwargs
        pass



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Natively supported in-context learning evaluation datasets."""

from llmfoundry.eval.datasets.in_context_learning_evaluation import (
    InContextLearningDataset,
    InContextLearningGenerationTaskWithAnswersDataset,
    InContextLearningLMTaskDataset,
    InContextLearningMultipleChoiceTaskDataset,
    InContextLearningSchemaTaskDataset,
    get_icl_task_dataloader,
)
from llmfoundry.eval.datasets.utils import (
    MultiTokenEOSCriteria,
    convert_tokens_to_tensors,
    get_continuation_span,
    get_fewshot_sample_idxs,
    make_padded_input,
    stop_sequences_criteria,
    strip_data,
    tokenizer_needs_prefix_space,
    trim_context,
)
from llmfoundry.registry import icl_datasets

icl_datasets.register(
    'multiple_choice',
    func=InContextLearningMultipleChoiceTaskDataset,
)
icl_datasets.register('schema', func=InContextLearningSchemaTaskDataset)
icl_datasets.register('language_modeling', func=InContextLearningLMTaskDataset)
icl_datasets.register(
    'generation_task_with_answers',
    func=InContextLearningGenerationTaskWithAnswersDataset,
)

__all__ = [
    'InContextLearningDataset',
    'InContextLearningGenerationTaskWithAnswersDataset',
    'InContextLearningLMTaskDataset',
    'InContextLearningMultipleChoiceTaskDataset',
    'InContextLearningSchemaTaskDataset',
    'get_icl_task_dataloader',
    'MultiTokenEOSCriteria',
    'strip_data',
    'tokenizer_needs_prefix_space',
    'trim_context',
    'get_continuation_span',
    'make_padded_input',
    'convert_tokens_to_tensors',
    'get_fewshot_sample_idxs',
    'stop_sequences_criteria',
]



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

# Copyright 2022 The HuggingFace Team. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
# Modified from https://github.com/huggingface/accelerate/blob/main/src/accelerate/big_modeling.py
from typing import Any, Callable, Optional

import torch
import torch.nn as nn
from torch.distributed._tensor import DTensor

__all__ = [
    'init_empty_weights',
    'init_on_device',
]


@contextmanager
def init_empty_weights(include_buffers: bool = False):
    """Meta initialization context manager.

    A context manager under which models are initialized with all parameters
    on the meta device, therefore creating an empty model. Useful when just
    initializing the model would blow the available RAM.

    Args:
        include_buffers (`bool`, *optional*, defaults to `False`): Whether or
            not to also put all buffers on the meta device while initializing.

    Example:
    ```python
    import torch.nn as nn

    # Initialize a model with 100 billions parameters in no time and without using any RAM.
    with init_empty_weights():
        tst = nn.Sequential(*[nn.Linear(10000, 10000) for _ in range(1000)])
    ```

    <Tip warning={true}>

    Any model created under this context manager has no weights. As such you can't do something like
    `model.to(some_device)` with it. To load weights inside your empty model, see [`load_checkpoint_and_dispatch`].

    </Tip>
    """
    with init_on_device(
        torch.device('meta'),
        include_buffers=include_buffers,
    ) as f:
        yield f


@contextmanager
def init_on_device(device: torch.device, include_buffers: bool = False):
    """Device initialization context manager.

    A context manager under which models are initialized with all parameters
    on the specified device.

    Args:
        device (`torch.device`): Device to initialize all parameters on.
        include_buffers (`bool`, *optional*, defaults to `False`): Whether or
            not to also put all buffers on the meta device while initializing.

    Example:
    ```python
    import torch.nn as nn

    with init_on_device(device=torch.device("cuda")):
        tst = nn.Liner(100, 100)  # on `cuda` device
    ```
    """
    old_register_parameter = nn.Module.register_parameter
    if include_buffers:
        old_register_buffer = nn.Module.register_buffer

    def register_empty_parameter(
        self: torch.nn.Module,
        name: str,
        param: Optional[torch.nn.Parameter],
    ):
        old_register_parameter(self, name, param)
        if param is not None:
            parameter = self._parameters[name]
            assert parameter is not None
            if isinstance(parameter, DTensor):
                self._parameters[name] = parameter.to(device)  # type: ignore
            else:
                param_cls = type(parameter)
                kwargs = parameter.__dict__
                self._parameters[name] = param_cls(
                    parameter.to(device),
                    requires_grad=parameter.requires_grad,  # requires grad is separate from the kwargs
                    **kwargs,
                )

    def register_empty_buffer(
        self: torch.nn.Module,
        name: str,
        tensor: Optional[torch.Tensor],
        persistent: bool = True,
    ):
        old_register_buffer(self, name, tensor, persistent=persistent)
        if tensor is not None:
            named_buffer = self._buffers[name]
            assert named_buffer is not None
            self._buffers[name] = named_buffer.to(device)

    # Patch tensor creation
    if include_buffers:
        tensor_constructors_to_patch = {
            torch_function_name: getattr(torch, torch_function_name)
            for torch_function_name in ['empty', 'zeros', 'ones', 'full']
        }
    else:
        tensor_constructors_to_patch = {}

    def patch_tensor_constructor(fn: Callable):

        def wrapper(*args: Any, **kwargs: Any):
            kwargs['device'] = device
            return fn(*args, **kwargs)

        return wrapper

    try:
        nn.Module.register_parameter = register_empty_parameter
        if include_buffers:
            nn.Module.register_buffer = register_empty_buffer
        for torch_function_name in tensor_constructors_to_patch.keys():
            setattr(
                torch,
                torch_function_name,
                patch_tensor_constructor(getattr(torch, torch_function_name)),
            )
        yield
    finally:
        nn.Module.register_parameter = old_register_parameter
        if include_buffers:
            nn.Module.register_buffer = old_register_buffer
        for torch_function_name, old_torch_function in tensor_constructors_to_patch.items(
        ):
            setattr(torch, torch_function_name, old_torch_function)



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Log tokens per expert for MegaBlocks MoE."""
from __future__ import annotations

import torch
from composer.core import Callback, State
from composer.loggers import Logger
from composer.utils import dist

__all__ = ['MegaBlocksMoE_TokPerExpert']


class MegaBlocksMoE_TokPerExpert(Callback):
    """Log tokens per expert for MegaBlocks MoE.

    To compute the load balancing loss, MegaBlocks caches information including `tokens_per_expert`
    (tpe). At the :attr:`.Event.BATCH_END` event this callback gets load_balancing_loss from
    MegaBlocks to get `tokens_per_expert` then logs statistics (<STAT>) of the number of tokens
    assigned to experts for each layer index (l_idx) under ``mb_moe/layer<l_idx>_<STAT>_tpe``.


    The tokens_per_expert statistics are logged by the :class:`.Logger` to the following keys as
    described below.

    +----------------------------------+-----------------------------------------------------------+
    | Key                              | Logged data                                               |
    +==================================+===========================================================+
    | `mb_moe/alllayer_min_tpe`        | Minimum tokens per expert across all layers               |
    +----------------------------------+-----------------------------------------------------------+
    | `mb_moe/alllayer_max_tpe`        | Maximum tokens per expert across all layers               |
    +----------------------------------+-----------------------------------------------------------+
    | `mb_moe/alllayer_median_tpe`     | Median tokens per expert across all layers                |
    +----------------------------------+-----------------------------------------------------------+
    | `mb_moe/alllayer_std_tpe`        | Standard deviation of tokens per expert across all layers |
    +----------------------------------+-----------------------------------------------------------+
    | `mb_moe/layer<l_idx>_min_tpe`    | Minimum tokens per expert at l_idx layer                  |
    +----------------------------------+-----------------------------------------------------------+
    | `mb_moe/layer<l_idx>_max_tpe`    | Maximum tokens per expert at l_idx layer                  |
    +----------------------------------+-----------------------------------------------------------+
    | `mb_moe/layer<l_idx>_median_tpe` | Median tokens per expert at l_idx layer                   |
    +----------------------------------+-----------------------------------------------------------+
    | `mb_moe/layer<l_idx>_std_tpe`    | Standard deviation of tokens per expert at l_idx layer    |
    +----------------------------------+-----------------------------------------------------------+

    Args:
        log_interval (int, optional): The interval on which to log (Default: 10).
        log_every_layer (bool, optional): Enable logging ever layer's statistics (True) or log
            only aggregate statistics (Default: False).
        all_reduce_stats (bool, optional): Enable aggregating statistics across gpus (True) or log
            statistics for GPU 0 (Default: False).
        normalize (bool, optional): Normalize token counts by total tokens (Default: True) or output
            raw token count (False). When normalize is True, the callback displays the fraction of
            unique tokens routed to each expert. When normalize is False, the callback displays the
            total number of tokens routed to each expert.
    """

    def __init__(
        self,
        log_interval: int = 10,
        log_every_layer: bool = False,
        all_reduce_stats: bool = False,
        normalize: bool = True,
    ):
        self.log_interval = log_interval
        self.log_every_layer = log_every_layer
        self.all_reduce_stats = all_reduce_stats
        self.normalize = normalize

        self.topk = None

    def fit_start(self, state: State, logger: Logger) -> None:
        if self.topk is None and self.normalize:
            try:
                from megablocks.layers.dmoe import dMoE
                from megablocks.layers.moe import MoE
            except:
                raise RuntimeError(
                    'Requirements for MegaBlocks not installed; see install instructions in `README.md`.',
                )
            for module in state.model.modules():
                if isinstance(module, (MoE, dMoE)):
                    self.topk = module.experts.args.moe_top_k
                    return

            raise RuntimeError(
                f'Callback not initialized correctly; self.topk not instantiated.',
            )

    def batch_end(self, state: State, logger: Logger) -> None:
        if state.timestamp.batch.value % self.log_interval == 0:
            try:
                from megablocks.layers.moe import get_load_balancing_loss
            except:
                raise RuntimeError(
                    'Requirements for MegaBlocks not installed; see install instructions in `README.md`.',
                )
            tokens_per_expert, _ = zip(*get_load_balancing_loss())

            tokens_per_expert = [
                tpe.clone().detach() for tpe in tokens_per_expert
            ]
            if self.all_reduce_stats:
                for tpe in tokens_per_expert:
                    dist.all_reduce(tpe)

            if self.normalize:
                tokens_per_expert = [
                    tpe / (tpe.sum() / self.topk) for tpe in tokens_per_expert
                ]

            all_tokens_per_expert = torch.concat(tokens_per_expert)

            min_tpe = all_tokens_per_expert.min().item()
            max_tpe = all_tokens_per_expert.max().item()
            median_tpe = all_tokens_per_expert.median().item()
            std_tpe = all_tokens_per_expert.float().std().item()

            log_info = {
                f'mb_moe/all_layers_min_tpe': min_tpe,
                f'mb_moe/all_layers_max_tpe': max_tpe,
                f'mb_moe/all_layers_median_tpe': median_tpe,
                f'mb_moe/all_layers_std_tpe': std_tpe,
            }

            if self.log_every_layer:
                for l_idx, tpe_layer in enumerate(tokens_per_expert):

                    min_tpe = tpe_layer.min().item()
                    max_tpe = tpe_layer.max().item()
                    median_tpe = tpe_layer.median().item()
                    std_tpe = tpe_layer.float().std().item()

                    log_info.update({
                        f'mb_moe/layer{l_idx}_min_tpe': min_tpe,
                        f'mb_moe/layer{l_idx}_max_tpe': max_tpe,
                        f'mb_moe/layer{l_idx}_median_tpe': median_tpe,
                        f'mb_moe/layer{l_idx}_std_tpe': std_tpe,
                    })

            logger.log_metrics(log_info)



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from llmfoundry.models.layers.attention import (
    GroupedQueryAttention,
    MultiheadAttention,
    MultiQueryAttention,
    attn_bias_shape,
    build_alibi_bias,
    build_attn_bias,
    check_alibi_support,
    flash_attn_fn,
    scaled_multihead_dot_product_attention,
)
from llmfoundry.models.layers.blocks import FusedNormAttentionNorm, MPTBlock
from llmfoundry.models.layers.custom_embedding import SharedEmbedding
from llmfoundry.models.layers.dmoe import DroplessMLP, LearnedRouter, dMoE
from llmfoundry.models.layers.fc import *
from llmfoundry.models.layers.ffn import MPTGLU, MPTMLP
from llmfoundry.models.layers.layer_builders import (
    build_attention_layer,
    build_fc,
    build_ffn,
    build_norm,
)
from llmfoundry.models.layers.norm import (
    LPLayerNorm,
    LPRMSNorm,
    RMSNorm,
    TritonRMSNorm,
    rms_norm,
)

__all__ = [
    'scaled_multihead_dot_product_attention',
    'flash_attn_fn',
    'MultiheadAttention',
    'MultiQueryAttention',
    'GroupedQueryAttention',
    'attn_bias_shape',
    'build_attn_bias',
    'build_alibi_bias',
    'check_alibi_support',
    'MPTBlock',
    'FusedNormAttentionNorm',
    'SharedEmbedding',
    'dMoE',
    'LearnedRouter',
    'DroplessMLP',
    'MPTMLP',
    'MPTGLU',
    'build_attention_layer',
    'build_ffn',
    'build_fc',
    'build_norm',
    'LPLayerNorm',
    'LPRMSNorm',
    'RMSNorm',
    'TritonRMSNorm',
    'rms_norm',
]



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import signal
import threading
from typing import Optional

from composer import Callback, Logger, State

log = logging.getLogger(__name__)


def _timeout(timeout: int):
    log.error(f'Timeout after {timeout} seconds of inactivity after fit_end.',)
    os.kill(os.getpid(), signal.SIGINT)


class RunTimeoutCallback(Callback):

    def __init__(
        self,
        timeout: int = 1800,
    ):
        self.timeout = timeout
        self.timer: Optional[threading.Timer] = None

    def _reset(self):
        if self.timer is not None:
            self.timer.cancel()
        self.timer = None

    def _timeout(self):
        self._reset()
        self.timer = threading.Timer(
            self.timeout,
            _timeout,
            [self.timeout],
        )
        self.timer.daemon = True
        self.timer.start()

    def fit_end(self, state: State, logger: Logger):
        del state
        del logger
        self._timeout()



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Implements a OpenAI chat and causal LM inference API wrappers."""

import logging
import os
import random
from time import sleep
from typing import TYPE_CHECKING, Any, Optional, Union

import torch
from composer.core.types import Batch
from composer.utils.import_helpers import MissingConditionalImportError
from omegaconf import DictConfig
from transformers import AutoTokenizer

log = logging.getLogger(__name__)

from llmfoundry.models.inference_api_wrapper.interface import \
    InferenceAPIEvalWrapper

__all__ = [
    'OpenAICausalLMEvalWrapper',
    'OpenAIChatAPIEvalWrapper',
    'OpenAIEvalInterface',
]

if TYPE_CHECKING:
    from openai.types.chat.chat_completion import ChatCompletion
    from openai.types.completion import Completion
    from openai.types.completion_choice import Logprobs

MAX_RETRIES = 10


class OpenAIEvalInterface(InferenceAPIEvalWrapper):

    def __init__(
        self,
        om_model_config: DictConfig,
        tokenizer: AutoTokenizer,
    ) -> None:
        super().__init__(om_model_config, tokenizer)
        try:
            import openai
        except ImportError as e:
            raise MissingConditionalImportError(
                extra_deps_group='openai',
                conda_package='openai',
                conda_channel='conda-forge',
            ) from e

        api_key = os.environ.get('OPENAI_API_KEY')
        base_url = om_model_config.get('base_url')
        if base_url is None:
            # Using OpenAI default, where the API key is required
            if api_key is None:
                raise ValueError(
                    'No OpenAI API Key found. Ensure it is saved as an environmental variable called OPENAI_API_KEY.',
                )

        else:
            # Using a custom base URL, where the API key may not be required
            log.info(
                f'Making request to custom base URL: {base_url}{"" if api_key is not None else " (no API key set)"}',
            )
            api_key = 'placeholder'  # This cannot be None

        self.client = openai.OpenAI(base_url=base_url, api_key=api_key)
        if 'version' in om_model_config:
            self.model_name = om_model_config['version']
        else:
            self.model_name = om_model_config['name']

    def generate_completion(self, prompt: str, num_tokens: int):
        raise NotImplementedError()

    def process_result(self, completion):  # pyright: ignore
        raise NotImplementedError()

    def get_next_token_logit_tensor(self, prompt: str, num_tokens: int = 1):
        completion = self.try_generate_completion(prompt, num_tokens)
        return self.process_result(completion)

    def try_generate_completion(self, prompt: str, num_tokens: int):
        try:
            from openai import APITimeoutError, RateLimitError
        except ImportError as e:
            raise MissingConditionalImportError(
                extra_deps_group='openai',
                conda_package='openai',
                conda_channel='conda-forge',
            ) from e
        tries = 0
        completion = None
        delay = 1
        while tries < MAX_RETRIES:
            tries += 1
            try:
                completion = self.generate_completion(prompt, num_tokens)
                break
            except RateLimitError as e:
                if 'You exceeded your current quota' in str(
                    e.message,
                ):  # pyright: ignore
                    raise e
                delay *= 2 * (1 + random.random())
                sleep(delay)
                continue
            except APITimeoutError as e:
                delay *= 2 * (1 + random.random())
                sleep(delay)
                continue

        return completion


class OpenAIChatAPIEvalWrapper(OpenAIEvalInterface):

    def __init__(
        self,
        om_model_config: DictConfig,
        tokenizer: AutoTokenizer,
    ) -> None:
        super().__init__(om_model_config, tokenizer)

        self.generate_completion = lambda prompt, num_tokens: self.client.chat.completions.create(
            model=self.model_name,
            messages=[{
                'role':
                    'system',
                'content':
                    om_model_config.get(
                        'system_role_prompt',
                        'Please complete the following text: ',
                    ),
            }, {
                'role': 'user',
                'content': prompt,
            }],
            max_tokens=num_tokens,
            temperature=0.0,
        )

    def retokenize(self, tokens: list[int], cont_idxs: list[int]):
        """Chat API will never respond with a word-initial space.

        If the continuation tokens begin with a word initial space, we need to
        re-tokenize with the space removed.
        """
        original_len = len(tokens)
        retokenized_continuation = self.tokenizer(
            self.tokenizer.decode(
                tokens[cont_idxs[0]:cont_idxs[-1] + 1],
            ).strip(),
        )['input_ids']

        # replace the original continuation with the retokenized continuation + padding
        padding = [tokens[-1]] * (
            len(tokens) - len(tokens[:cont_idxs[0]] + retokenized_continuation)
        )
        tokens = tokens[:cont_idxs[0]] + retokenized_continuation + padding

        if len(tokens) > original_len:
            # this only happens if we were already at max seq len and the continuation got LARGER
            tokens = tokens[-original_len:]
            cont_idxs = list(
                range(
                    original_len - len(retokenized_continuation),
                    original_len,
                ),
            )
        else:
            cont_idxs = list(
                range(
                    cont_idxs[0],
                    cont_idxs[0] + len(retokenized_continuation),
                ),
            )
        return torch.tensor(tokens), torch.tensor(cont_idxs)

    def rebatch(self, batch: Batch):
        """Chat API tokenization has different behavior than GPT3.

        Model responses will never begin with spaces even if the continuation is
        expected to, so we need to retokenize the input to account for that.
        """
        new_batch: dict[str, Union[list[torch.Tensor], torch.Tensor]] = {
            'input_ids': [],
            'continuation_indices': [],
            'labels': [],
        }
        for tokens, cont_idxs in zip(
            batch['input_ids'],
            batch['continuation_indices'],
        ):
            tokens, cont_idxs = self.retokenize(
                tokens.tolist(),
                cont_idxs.tolist(),
            )

            assert isinstance(new_batch['input_ids'], list)
            new_batch['input_ids'].append(tokens)
            assert isinstance(new_batch['labels'], list)
            new_batch['labels'].append(tokens)
            assert isinstance(new_batch['continuation_indices'], list)
            new_batch['continuation_indices'].append(cont_idxs)

        new_batch.update({
            k: torch.stack(new_batch[k])  # pyright: ignore
            for k in ['input_ids', 'labels']
        })

        new_batch.update({k: v for k, v in batch.items() if k not in new_batch})

        return new_batch

    def eval_forward(self, batch: Batch, outputs: Optional[Any] = None):
        # Override the base class because Chat's API always strips spacing from model outputs resulting in different tokens
        # than what the continuation would expect.
        # Get around this issue by retokenizing the batch to remove spacing from the continuation as well as
        # decoding the whole continuation at once.
        padding_tok = self.tokenizer.pad_token_id if self.tokenizer.pad_token_id else self.tokenizer.eos_token_id
        output_logits_batch = []
        batch = self.rebatch(batch)
        for tokens, cont_idxs in zip(
            batch['input_ids'],
            batch['continuation_indices'],
        ):

            seqlen = tokens.shape[0]
            tokens = tokens.tolist()
            cont_idxs = cont_idxs.tolist()
            expected_cont_tokens = tokens[cont_idxs[0]:cont_idxs[-1] + 1]
            output_logits = torch.nn.functional.one_hot(
                torch.tensor(tokens[1:cont_idxs[0]]),
                num_classes=len(self.tokenizer),
            )

            prompt = self.tokenizer.decode(tokens[:cont_idxs[0]])
            next_logit_tensor = self.get_next_token_logit_tensor(
                prompt,
                num_tokens=len(expected_cont_tokens),
            )

            if next_logit_tensor is not None:
                output_logits = torch.cat([output_logits, next_logit_tensor])
            padding = torch.nn.functional.one_hot(
                torch.full((seqlen - output_logits.shape[0],), padding_tok),
                num_classes=len(self.tokenizer),
            )
            output_logits = torch.cat([output_logits, padding])
            output_logits_batch.append(output_logits)

        return torch.stack(output_logits_batch).to(batch['input_ids'].device)

    def process_result(self, completion: Optional['ChatCompletion']):
        if completion is None:
            raise ValueError("Couldn't generate model output")

        if len(completion.choices) > 0:
            tensors = []
            for t in self.tokenizer(
                completion.choices[0].message.content,
            )['input_ids']:
                # Not real logprobs
                tensor = torch.tensor([0] * (len(self.tokenizer)))
                tensor[t] = 1.0
                tensors.append(tensor)

            if len(tensors) == 0:
                return None
            return torch.stack(tensors)
        else:
            # the model sometimes stops early even though we are still requesting tokens!
            # not sure if there's a fix
            return None


class OpenAICausalLMEvalWrapper(OpenAIEvalInterface):

    def __init__(
        self,
        om_model_config: DictConfig,
        tokenizer: AutoTokenizer,
    ) -> None:
        super().__init__(om_model_config, tokenizer)
        self.generate_completion = lambda prompt, num_tokens: self.client.completions.create(
            model=self.model_name,
            prompt=prompt,
            max_tokens=num_tokens,
            logprobs=5,
            temperature=0.0,
        )

    def process_result(self, completion: Optional['Completion']):
        if completion is None:
            raise ValueError("Couldn't generate model output")

        if TYPE_CHECKING:
            assert isinstance(completion, Completion)
            assert isinstance(completion.choices[0].logprobs, Logprobs)
            assert isinstance(completion.choices[0].logprobs.top_logprobs, list)

        if len(completion.choices[0].logprobs.top_logprobs[0]) > 0:
            # Construct tensor of shape (vocab_size,) with logprobs for each token
            tokenizer_logprobs = dict(
                completion.choices[0].logprobs.top_logprobs[0],
            )
            tensor = torch.tensor([min(tokenizer_logprobs.values()) - 1] *
                                  (len(self.tokenizer)))
            for k in tokenizer_logprobs:
                encoding = self.tokenizer(k)['input_ids']
                tensor[encoding[0]] = tokenizer_logprobs[k]
            return tensor
        else:
            # the model sometimes stops early even though we are still requesting tokens!
            # not sure if there's a fix
            return None



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0
from typing import Any, Callable, Iterable, Union

from composer.core import Algorithm, Callback, DataSpec
from composer.loggers import LoggerDestination
from composer.models import ComposerModel
from composer.optim import ComposerScheduler
from torch.distributed.checkpoint import LoadPlanner, SavePlanner
from torch.distributed.tensor.parallel.style import ParallelStyle
from torch.optim import Optimizer
from torch.utils.data import DataLoader as TorchDataloader
from torch.utils.data import Dataset
from torchmetrics import Metric
from transformers import PreTrainedTokenizerBase

from llmfoundry.interfaces import CallbackWithConfig
from llmfoundry.layers_registry import (
    attention_classes,
    attention_implementations,
    fcs,
    ffns,
    ffns_with_megablocks,
    ffns_with_norm,
    module_init_fns,
    norms,
    param_init_fns,
)
from llmfoundry.utils.registry_utils import create_registry

_loggers_description = (
    """The loggers registry is used to register classes that implement the LoggerDestination interface.

    These classes are used to log data from the training loop, and will be passed to the loggers arg of the Trainer. The loggers
    will be constructed by directly passing along the specified kwargs to the constructor. See loggers/ for examples.

    Args:
        kwargs (Dict[str, Any]): The kwargs to pass to the LoggerDestination constructor.

    Returns:
        LoggerDestination: The logger destination.
    """
)
loggers = create_registry(
    'llmfoundry',
    'loggers',
    generic_type=type[LoggerDestination],
    entry_points=True,
    description=_loggers_description,
)

_callbacks_description = (
    """The callbacks registry is used to register classes that implement the Callback interface.

    These classes are used to interact with the Composer event system, and will be passed to the callbacks arg of the Trainer.
    The callbacks will be constructed by directly passing along the specified kwargs to the constructor. See callbacks/ for examples.

    Args:
        kwargs (Dict[str, Any]): The kwargs to pass to the Callback constructor.

    Returns:
        Callback: The callback.
    """
)
callbacks = create_registry(
    'llmfoundry',
    'callbacks',
    generic_type=type[Callback],
    entry_points=True,
    description=_callbacks_description,
)

_callbacks_with_config_description = (
    """The callbacks_with_config registry is used to register classes that implement the CallbackWithConfig interface.

    These are the same as the callbacks registry, except that they additionally take the full training config as an argument to their constructor.
    See callbacks/ for examples.

    Args:
        config (DictConfig): The training config.
        kwargs (Dict[str, Any]): The kwargs to pass to the Callback constructor.

    Returns:
        Callback: The callback.
    """
)
callbacks_with_config = create_registry(
    'llmfoundry',
    'callbacks_with_config',
    generic_type=type[CallbackWithConfig],
    entry_points=True,
    description=_callbacks_with_config_description,
)

_optimizers_description = (
    """The optimizers registry is used to register classes that implement the Optimizer interface.

    The optimizer will be passed to the optimizers arg of the Trainer. The optimizer will be constructed by directly passing along the
    specified kwargs to the constructor, along with the model parameters. See optim/ for examples.

    Args:
        params (Iterable[torch.nn.Parameter]): The model parameters.
        kwargs (Dict[str, Any]): The kwargs to pass to the Optimizer constructor.

    Returns:
        Optimizer: The optimizer.
    """
)
optimizers = create_registry(
    'llmfoundry',
    'optimizers',
    generic_type=type[Optimizer],
    entry_points=True,
    description=_optimizers_description,
)

_algorithms_description = (
    """The algorithms registry is used to register classes that implement the Algorithm interface.

    The algorithm will be passed to the algorithms arg of the Trainer. The algorithm will be constructed by directly passing along the
    specified kwargs to the constructor. See algorithms/ for examples.

    Args:
        kwargs (Dict[str, Any]): The kwargs to pass to the Algorithm constructor.

    Returns:
        Algorithm: The algorithm.
    """
)
algorithms = create_registry(
    'llmfoundry',
    'algorithms',
    generic_type=type[Algorithm],
    entry_points=True,
    description=_algorithms_description,
)

_schedulers_description = (
    """The schedulers registry is used to register classes that implement the ComposerScheduler interface.

    The scheduler will be passed to the schedulers arg of the Trainer. The scheduler will be constructed by directly passing along the
    specified kwargs to the constructor. See optim/ for examples.

    Args:
        kwargs (Dict[str, Any]): The kwargs to pass to the ComposerScheduler constructor.

    Returns:
        ComposerScheduler: The scheduler.
    """
)
schedulers = create_registry(
    'llmfoundry',
    'schedulers',
    generic_type=type[ComposerScheduler],
    entry_points=True,
    description=_schedulers_description,
)

_tokenizers_description = (
    'The tokenizers registry is used to register tokenizers that implement the transformers.PreTrainedTokenizerBase interface. '
    +
    'The tokenizer will be passed to the build_dataloader() and build_composer_model() methods in train.py.'
)
tokenizers = create_registry(
    'llmfoundry',
    'tokenizers',
    generic_type=type[PreTrainedTokenizerBase],
    entry_points=True,
    description=_tokenizers_description,
)

_models_description = (
    """The models registry is used to register classes that implement the ComposerModel interface.

    The model constructor should accept a PreTrainedTokenizerBase named `tokenizer`, and the rest of its constructor kwargs.
    See models/ for examples.

    Args:
        tokenizer (PreTrainedTokenizerBase): The tokenizer.
        kwargs (Dict[str, Any]): The kwargs to pass to the Composer

    Returns:
        ComposerModel: The model.
    """
)
models = create_registry(
    'llmfoundry',
    'models',
    generic_type=type[ComposerModel],
    entry_points=True,
    description=_models_description,
)

_dataloaders_description = (
    """The dataloaders registry is used to register functions that create a DataSpec given a config.

    The function should take a PreTrainedTokenizerBase, a device batch size, and the rest of its constructor kwargs,
    and return a DataSpec. See data/ for examples.

    Args:
        tokenizer (PreTrainedTokenizerBase): The tokenizer
        device_batch_size (Union[int, float]): The device batch size.
        kwargs (Dict[str, Any]): The kwargs to pass to the builder function.

    Returns:
        DataSpec: The dataspec.
    """
)
dataloaders = create_registry(
    'llmfoundry',
    'dataloaders',
    generic_type=Callable[
        ...,
        DataSpec],  # The arguments to the dataloader may vary depending on the contents of the config.
    entry_points=True,
    description=_dataloaders_description,
)

_dataset_replication_validators_description = (
    """The dataset_replication_validators registry is used to register functions that validate replication factor.

    The function should return the replication factor and the dataset device batch size. See data/ for examples.

    Args:
        cfg (DictConfig): The dataloader config.
        tokenizer (PreTrainedTokenizerBase): The tokenizer
        device_batch_size (Union[int, float]): The device batch size.

    Returns:
        replication_factor (int): The replication factor for dataset.
        dataset_batch_size (int): The dataset device batch size.
    """
)
dataset_replication_validators = create_registry(
    'llmfoundry',
    'dataset_replication_validators',
    generic_type=Callable[
        [dict[str, Any], PreTrainedTokenizerBase, Union[int, float]],
        tuple[int, int]],
    entry_points=True,
    description=_dataset_replication_validators_description,
)

_collators_description = (
    """The collators registry is used to register functions that create the collate function for the DataLoader.

    See data/ for examples.

    Args:
        cfg (DictConfig): The dataloader config.
        tokenizer (PreTrainedTokenizerBase): The tokenizer
        dataset_batch_size (Union[int, float]): The dataset device batch size.

    Returns:
        collate_fn  (Any): The collate function.
        dataloader_batch_size (int): The batch size for dataloader. In case of packing, this might be the packing ratio times the dataset device batch size.
    """
)
collators = create_registry(
    'llmfoundry',
    'collators',
    generic_type=Callable[[dict[str, Any], PreTrainedTokenizerBase, int],
                          tuple[Any, int]],
    entry_points=True,
    description=_collators_description,
)

_data_specs_description = (
    """The data_specs registry is used to register functions that create a DataSpec given a dataloader.

    See data/ for examples.

    Args:
        dl (Union[Iterable, TorchDataloader): The dataloader.
        dataset_cfg (DictConfig): The dataset config.

    Returns:
        dataspec (DataSpec): The dataspec.
    """
)
data_specs = create_registry(
    'llmfoundry',
    'data_specs',
    generic_type=Callable[[Union[Iterable, TorchDataloader], dict[str, Any]],
                          DataSpec],
    entry_points=True,
    description=_data_specs_description,
)

_metrics_description = (
    """The metrics registry is used to register classes that implement the torchmetrics.Metric interface.

    The metric will be passed to the metrics arg of the Trainer. The metric will be constructed by directly passing along the
    specified kwargs to the constructor. See metrics/ for examples.

    Args:
        kwargs (Dict[str, Any]): The kwargs to pass to the Metric constructor.

    Returns:
        Metric: The metric.
    """
)
metrics = create_registry(
    'llmfoundry',
    'metrics',
    generic_type=type[Metric],
    entry_points=True,
    description=_metrics_description,
)

_icl_datasets_description = (
    """The ICL datasets registry is used to register classes that implement the InContextLearningDataset interface.

    The dataset will be constructed along with an Evaluator. The dataset will be constructed by directly passing along the
    specified kwargs to the constructor. See eval/ for examples.

    Args:
        kwargs (Dict[str, Any]): The kwargs to pass to the Dataset constructor.

    Returns:
        InContextLearningDataset: The dataset.
    """
)
icl_datasets = create_registry(
    'llmfoundry',
    'icl_datasets',
    # TODO: Change type from Dataset to
    # llmfoundry.eval.InContextLearningDataset.
    # Using ICL dataset here introduces a circular import dependency between
    # the registry and eval packages right now, thus needs some refactoring.
    generic_type=type[Dataset],
    entry_points=True,
    description=_icl_datasets_description,
)

_config_transforms_description = (
    """The config_transforms registry is used to register functions that transform the training config

    The config will be transformed before it is used anywhere else. Note: By default ALL registered transforms will be applied to the train config
    and NONE to the eval config. Each transform should return the modified config. See utils/config_utils.py for examples.

    Args:
        cfg (Dict[str, Any]): The training config.

    Returns:
        cfg (Dict[str, Any]): The modified training config.
    """
)
config_transforms = create_registry(
    'llmfoundry',
    'config_transforms',
    generic_type=Callable[[dict[str, Any]], dict[str, Any]],
    entry_points=True,
    description=_config_transforms_description,
)

_load_planners_description = (
    """The load_planners registry is used to register classes that implement the LoadPlanner interface.

    The LoadPlanner will be passed as part of the FSDP config arg of the Trainer. It will be used to load distributed checkpoints.

    Returns:
        LoadPlanner: The load planner.
    """
)

load_planners = create_registry(
    'llmfoundry',
    'load_planners',
    generic_type=type[LoadPlanner],
    entry_points=True,
    description=_load_planners_description,
)

_save_planners_description = (
    """The save_planners registry is used to register classes that implement the SavePlanner interface.

    The savePlanner will be passed as part of the FSDP config arg of the Trainer. It will be used to save distributed checkpoints.

    Returns:
        SavePlanner: The save planner.
    """
)

save_planners = create_registry(
    'llmfoundry',
    'save_planners',
    generic_type=type[SavePlanner],
    entry_points=True,
    description=_save_planners_description,
)

_tp_strategies_description = (
    """The tp_strategies registry is used to register strategies for tensor parallelism.

    Args:
        model (ComposerModel): The model.

    Returns:
        layer_plan (Dict[str, ParallelStyle]): The plan used to parallelize the model.
        model (ComposerModel): The model.
    """
)

tp_strategies = create_registry(
    'llmfoundry',
    'tp_strategies',
    generic_type=Callable[[ComposerModel], dict[str, ParallelStyle]],
    entry_points=True,
    description=_tp_strategies_description,
)

__all__ = [
    'loggers',
    'callbacks',
    'callbacks_with_config',
    'optimizers',
    'algorithms',
    'schedulers',
    'tokenizers',
    'models',
    'dataset_replication_validators',
    'collators',
    'data_specs',
    'metrics',
    'dataloaders',
    'norms',
    'param_init_fns',
    'module_init_fns',
    'ffns',
    'ffns_with_norm',
    'ffns_with_megablocks',
    'attention_classes',
    'attention_implementations',
    'fcs',
    'icl_datasets',
    'config_transforms',
    'load_planners',
    'save_planners',
    'tp_strategies',
]



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import contextlib
import copy
import functools
import logging
import os
import re
import warnings
from collections import OrderedDict
from typing import Any, ContextManager, Iterable, Optional, Union

import torch
from composer.core import Algorithm, Callback, Evaluator
from composer.loggers import LoggerDestination
from composer.models import ComposerModel
from composer.optim.scheduler import ComposerScheduler
from composer.utils import dist
from omegaconf import DictConfig
from omegaconf import OmegaConf as om
from torch.distributed.checkpoint import LoadPlanner, SavePlanner
from torch.distributed.tensor.parallel.style import ParallelStyle
from torch.optim.optimizer import Optimizer
from torchmetrics import Metric
from transformers import AutoTokenizer, PreTrainedTokenizerBase

from llmfoundry import registry
from llmfoundry.callbacks import EvalGauntlet
from llmfoundry.data.dataloader import build_dataloader
from llmfoundry.eval.datasets.in_context_learning_evaluation import (
    get_icl_task_dataloader,
)
from llmfoundry.utils.config_utils import to_dict_container, to_list_container
from llmfoundry.utils.registry_utils import construct_from_registry
from llmfoundry.utils.warnings import experimental_function

log = logging.getLogger(__name__)

__all__ = [
    'build_algorithm',
    'build_callback',
    'build_evaluators',
    'build_icl_data_and_gauntlet',
    'build_icl_evaluators',
    'build_logger',
    'build_optimizer',
    'build_scheduler',
    'build_tokenizer',
    'build_composer_model',
    'build_metric',
    'build_tp_strategies',
]


def build_evaluators(
    eval_loader_config: Optional[Union[dict[str, Any], list[dict[str, Any]]]],
    icl_tasks_config: Optional[Union[str, list[dict[str, Any]]]],
    eval_gauntlet_config: Optional[Union[str, dict[str, Any]]],
    *,
    tokenizer: PreTrainedTokenizerBase,
    device_eval_batch_size: Union[int, float],
    icl_seq_len: int,
    icl_subset_num_batches: Optional[int],
) -> tuple[list[Evaluator], list[str], Optional[EvalGauntlet]]:

    evaluators = []
    if eval_loader_config is not None:
        evaluators = build_eval_loaders(
            eval_loader_config,
            tokenizer,
            device_eval_batch_size,
        )

    logger_keys = []
    eval_gauntlet_callback = None
    if icl_tasks_config is not None:
        if not isinstance(device_eval_batch_size, int):
            raise ValueError(
                'device_eval_batch_size should be an int for icl tasks.',
            )
        icl_evaluators, logger_keys, eval_gauntlet_callback = build_icl_data_and_gauntlet(
            icl_tasks_config,
            eval_gauntlet_config,
            tokenizer,
            device_eval_batch_size,
            icl_seq_len,
            icl_subset_num_batches,
        )
        evaluators.extend(icl_evaluators)

    return evaluators, logger_keys, eval_gauntlet_callback


def build_eval_loaders(
    eval_loader_config: Union[dict[str, Any], list[dict[str, Any]]],
    tokenizer: PreTrainedTokenizerBase,
    device_eval_batch_size: Union[int, float],
) -> list[Evaluator]:
    evaluators: list[Evaluator] = []
    if isinstance(eval_loader_config, list):
        eval_configs = eval_loader_config
        is_multi_eval = True
    elif isinstance(eval_loader_config, dict):
        eval_configs = [eval_loader_config]
        is_multi_eval = False
    else:
        raise ValueError(
            f'Got invalid type for eval_loader_config: {type(eval_loader_config)}, {eval_loader_config=}',
        )

    for eval_config in eval_configs:
        label = eval_config.pop('label') if is_multi_eval else None
        eval_dataloader = build_dataloader(
            eval_config,
            tokenizer,
            device_eval_batch_size,
        )
        eval_loader: Evaluator = Evaluator(
            label=f'eval/{label}' if is_multi_eval else 'eval',
            dataloader=eval_dataloader,
            # Load the eval data to fail fast. metrics will get added
            # later in add_metrics_to_eval_loaders, after the model is loaded
            metric_names=[],
            device_eval_microbatch_size=device_eval_batch_size,
        )
        evaluators.append(eval_loader)
    return evaluators


def add_metrics_to_eval_loaders(
    evaluators: list[Evaluator],
    metric_names: list[str],
) -> list[Evaluator]:
    eval_loaders, other_evaluators = [], []
    for evaluator in evaluators:
        if evaluator.metric_names == []:
            evaluator.metric_names = metric_names
            eval_loaders.append(evaluator)
        else:
            other_evaluators.append(evaluator)

    # Put the base eval_loaders first
    return eval_loaders + other_evaluators


def build_icl_data_and_gauntlet(
    icl_tasks_config: Union[str, list[dict[str, Any]]],
    eval_gauntlet_config: Optional[Union[str, dict[str, Any]]],
    tokenizer: PreTrainedTokenizerBase,
    device_eval_batch_size: int,
    icl_seq_len: int,
    icl_subset_num_batches: Optional[int] = None,
) -> tuple[list[Evaluator], list[str], Optional[EvalGauntlet]]:
    icl_evaluators, logger_keys = build_icl_evaluators(
        icl_tasks_config,
        tokenizer,
        icl_seq_len,
        device_eval_batch_size,
        icl_subset_num_batches=icl_subset_num_batches,
    )
    eval_gauntlet_cb = None
    if eval_gauntlet_config is not None:
        if isinstance(eval_gauntlet_config, str):
            with open(eval_gauntlet_config, 'r') as icl_f:
                eval_gauntlet_cfg = om.load(icl_f)
                assert isinstance(eval_gauntlet_cfg, DictConfig)
            eval_gauntlet = to_dict_container(
                eval_gauntlet_cfg['eval_gauntlet'],
            )
        elif isinstance(eval_gauntlet_config, dict):  # pyright: ignore
            eval_gauntlet = eval_gauntlet_config
        else:
            raise ValueError(
                f'Got invalid type for eval_gauntlet_config: {type(eval_gauntlet_config)}',
            )
        eval_gauntlet['logger_keys'] = logger_keys
        eval_gauntlet['benchmark_sizes'] = {
            e.label: e.dataloader.num_samples for e in icl_evaluators
        }
        eval_gauntlet_cb = EvalGauntlet(**eval_gauntlet)
    return icl_evaluators, logger_keys, eval_gauntlet_cb


def build_load_planner(name: str, **kwargs: Any) -> LoadPlanner:
    """Builds a load planner from the registry.

    Args:
        name (str): Name of the load planner to build.
        kwargs (Any): Other relevant keyword arguments.

    Returns:
        LoadPlanner: The load planner.
    """
    return construct_from_registry(
        name=name,
        registry=registry.load_planners,
        partial_function=True,
        pre_validation_function=LoadPlanner,
        post_validation_function=None,
        kwargs=kwargs,
    )


def build_save_planner(name: str, **kwargs: Any) -> SavePlanner:
    """Builds a save planner from the registry.

    Args:
        name (str): Name of the save planner to build.
        kwargs (Any): Other relevant keyword arguments.

    Returns:
        savePlanner: The save planner.
    """
    return construct_from_registry(
        name=name,
        registry=registry.save_planners,
        partial_function=True,
        pre_validation_function=SavePlanner,
        post_validation_function=None,
        kwargs=kwargs,
    )


def build_composer_model(
    name: str,
    cfg: dict[str, Any],
    tokenizer: PreTrainedTokenizerBase,
    init_context: Optional[ContextManager] = None,
    master_weights_dtype: Optional[str] = None,
) -> ComposerModel:
    """Builds a ComposerModel from the registry.

    Args:
        name (str): Name of the model to build.
        cfg (DictConfig): Configuration for the model.
        tokenizer (PreTrainedTokenizerBase): Tokenizer to use.
        init_context (Optional[ContextManager], optional): Context manager to use for initialization. Defaults to None.
        master_weights_dtype (Optional[str], optional): Master weights dtype. Defaults to None.

    Returns:
        ComposerModel: _description_
    """
    if init_context is None:
        init_context = contextlib.nullcontext()

    with init_context:
        model = construct_from_registry(
            name=name,
            registry=registry.models,
            pre_validation_function=ComposerModel,
            post_validation_function=None,
            kwargs={
                **cfg,
                'tokenizer': tokenizer,
            },
        )

    str_dtype_to_torch_dtype = {
        'f16': torch.float16,
        'float16': torch.float16,
        'bf16': torch.bfloat16,
        'bfloat16': torch.bfloat16,
    }

    if master_weights_dtype is not None:
        if master_weights_dtype not in str_dtype_to_torch_dtype:
            raise ValueError(
                f'Invalid master_weights_dtype: {master_weights_dtype}. ' +
                f'Valid options are: {list(str_dtype_to_torch_dtype.keys())}.',
            )
        dtype = str_dtype_to_torch_dtype[master_weights_dtype]
        model = model.to(dtype=dtype)

    return model


def build_callback(
    name: str,
    kwargs: Optional[dict[str, Any]] = None,
    train_config: Any = None,
) -> Callback:
    """Builds a callback from the registry."""
    registry_to_use = registry.callbacks
    if name in registry.callbacks_with_config:
        if kwargs is None:
            kwargs = {}
        if 'train_config' in kwargs:
            raise ValueError(
                f'`train_config` is a reserved keyword for callbacks with config. Please remove it from the kwargs.',
            )
        kwargs['train_config'] = copy.deepcopy(train_config)
        registry_to_use = registry.callbacks_with_config

    return construct_from_registry(
        name=name,
        registry=registry_to_use,
        partial_function=True,
        pre_validation_function=Callback,
        post_validation_function=None,
        kwargs=kwargs,
    )


def build_logger(
    name: str,
    kwargs: Optional[dict[str, Any]] = None,
) -> LoggerDestination:
    """Builds a logger from the registry."""
    return construct_from_registry(
        name=name,
        registry=registry.loggers,
        partial_function=True,
        pre_validation_function=LoggerDestination,
        post_validation_function=None,
        kwargs=kwargs,
    )


def build_algorithm(
    name: str,
    kwargs: Optional[dict[str, Any]] = None,
) -> Algorithm:
    """Builds an algorithm from the registry."""
    return construct_from_registry(
        name=name,
        registry=registry.algorithms,
        partial_function=True,
        pre_validation_function=Algorithm,
        post_validation_function=None,
        kwargs=kwargs,
    )


def build_metric(name: str, kwargs: Optional[dict[str, Any]] = None) -> Metric:
    """Builds a metric from the registry."""
    return construct_from_registry(
        name=name,
        registry=registry.metrics,
        partial_function=True,
        pre_validation_function=Metric,
        post_validation_function=None,
        kwargs=kwargs,
    )


def _extract_param_groups(
    model: torch.nn.Module,
    optimizer_config: Optional[dict[str, Any]] = None,
) -> Union[Iterable[torch.Tensor], Iterable[dict[str, Any]]]:
    """Extracts parameter groups defined in the optimizer config.

    The optimizer_config defines the optimizer args. It can additionally have key
    `disable_grad` which is a string or list of strings. If a string matches a
    parameter name, then that parameter will have `requires_grad=False`. This is
    useful for freezing parameters. It can additionally have a key
    `param_groups` which is a list of dicts. In this dict, key `param_str_match`
    defines a string; if a parameter name contains this string, then it will be
    in this parameter group. This is useful for grouping parameters together.
    The dict can also contain any other key that is a valid optimizer arg.
    Note: to handle name overlap conflicts, params are assigned to parameter
    groups and added to `param_groups` in the order that `param_str_match` appear
    in `param_groups`.

    Usage
    To disable gradient for all parameters that contain the string "norm" or "bias":
    ```
    optimizer_config: {
        "name": "decoupled_lionw",
        "lr": 1e-3,
        "weight_decay": 1e-2,
        "betas": [0.9, 0.999],
        "eps": 1e-8,
        "disable_grad": ["norm", "bias"]
    }
    ```

    To create and modify the optimizer parameters for all parameters that contain
    the string "norm" and "bias" separately:
    ```
    optimizer_config: {
        "name": "decoupled_lionw",
        "lr": 1e-3,
        "weight_decay": 1e-2,
        "betas": [0.9, 0.999],
        "eps": 1e-8,
        "param_groups": [
            {
                "param_str_match": "norm",
                "lr": 1e-4,
                "weight_decay": 0.0,
            },
            {
                "param_str_match": "bias",
                "lr": 5e-4,
                "weight_decay": 0.0,
            },
        ],
    }
    ```

    Args:
        model (torch.nn.Module): model to extract parameters from
        optimizer_config (Dict[str, Any]): optimizer config

    Returns:
        Union[Iterable[torch.Tensor], Iterable[Dict[str, Any]]]: an iterable of
            torch.Tensor's or dict's. Specifies what Tensors should be optimized
            and their param groupings.
    """
    if optimizer_config is None:
        return model.parameters()

    if 'disable_grad' in optimizer_config.keys():
        str_matches = optimizer_config.pop('disable_grad')
        if isinstance(str_matches, str):
            str_matches = [str_matches]
        for str_match in str_matches:
            for n, p in model.named_parameters():
                if re.search(str_match, n):
                    p.requires_grad = False
                    log.debug(f'Setting `{n}.requires_grad = False`.')

    param_groups_config = optimizer_config.pop('param_groups', None)
    if param_groups_config is not None:
        params = []
        param_dict = OrderedDict((n, p) for n, p in model.named_parameters())

        log.debug(f'Default optimizer settings: {optimizer_config}.')
        for param_group_config in param_groups_config:
            str_match = param_group_config.pop('param_str_match')
            filter_fn = functools.partial(re.search, str_match)
            param_names = [n for n in param_dict.keys() if filter_fn(n)]
            group_params = {'params': [param_dict.pop(n) for n in param_names]}
            group_params.update(param_group_config)

            log.debug(
                f'Creating optimizer param_group with parameters: {param_names} ' +\
                f'(extracted using {str_match=}). The param_group optimizer ' +\
                f'setting overrides are: {param_group_config}.')

            params.append(group_params)

        params.insert(0, {'params': param_dict.values()})
        return params

    return model.parameters()


def build_optimizer(
    model: torch.nn.Module,
    name: str,
    optimizer_config: dict[str, Any],
) -> Optimizer:

    params = _extract_param_groups(model, optimizer_config)
    kwargs = {**optimizer_config}

    if 'params' in kwargs:
        raise ValueError(
            'The `params` will be automatically extracted from the model and ' +
            'optimizer config. Please remove it from the optimizer config kwargs.',
        )

    kwargs['params'] = params
    return construct_from_registry(
        name=name,
        registry=registry.optimizers,
        partial_function=True,
        pre_validation_function=Optimizer,
        post_validation_function=None,
        kwargs=kwargs,
    )


def build_scheduler(
    name: str,
    scheduler_config: Optional[dict[str, Any]] = None,
) -> ComposerScheduler:
    return construct_from_registry(
        name=name,
        registry=registry.schedulers,
        partial_function=True,
        pre_validation_function=ComposerScheduler,
        post_validation_function=None,
        kwargs=scheduler_config,
    )


def build_tokenizer(
    tokenizer_name: str,
    tokenizer_kwargs: dict[str, Any],
) -> PreTrainedTokenizerBase:
    os.environ['TRANSFORMERS_NO_ADVISORY_WARNINGS'] = '1'
    os.environ['TOKENIZERS_PARALLELISM'] = 'false'

    signal_file_path = dist.get_node_signal_file_name()

    if dist.is_available() and dist.is_initialized(
    ) and dist.get_world_size() > 1:
        # Make sure the tokenizer files are downloaded and cached first by local rank 0
        with dist.local_rank_zero_download_and_wait(signal_file_path):
            pass

    if tokenizer_name in registry.tokenizers:
        tokenizer = construct_from_registry(
            name=tokenizer_name,
            registry=registry.tokenizers,
            partial_function=True,
            pre_validation_function=PreTrainedTokenizerBase,
            post_validation_function=None,
            kwargs=tokenizer_kwargs,
        )
    else:
        tokenizer = AutoTokenizer.from_pretrained(
            tokenizer_name,
            **tokenizer_kwargs,
        )

        # HuggingFace does not respect the model_max_length kwarg, and overrides it with
        # min(kwargs['model_max_length'], original_config['model_max_length']), so we
        # explicitly set it here
        tokenizer.model_max_length = tokenizer_kwargs.get(
            'model_max_length',
            int(1e30),
        )

    if not hasattr(tokenizer, 'eos_token') or tokenizer.eos_token is None:
        raise ValueError(
            f'The tokenizer {tokenizer_name} must have an eos_token.',
        )

    if dist.is_available() and dist.is_initialized(
    ) and dist.get_world_size() > 1:
        if dist.get_local_rank() == 0:
            with open(signal_file_path, 'wb') as f:
                f.write(b'local_rank0_completed_tokenizer_setup')

        dist.barrier()

        if dist.get_local_rank() == 0:
            os.remove(signal_file_path)

    return tokenizer


def build_icl_evaluators(
    icl_tasks: Union[str, list[dict[str, Any]]],
    tokenizer: PreTrainedTokenizerBase,
    default_max_seq_len: int,
    default_batch_size: int,
    destination_dir: Optional[str] = None,
    icl_subset_num_batches: Optional[int] = None,
) -> tuple[list[Evaluator], list[str]]:
    if destination_dir is None:
        destination_dir = os.getcwd()

    evaluators = []
    logger_keys = []

    icl_tasks_list = None
    if isinstance(icl_tasks, str):
        log.info(f'Extracting ICL task config from path: {icl_tasks}')
        with open(icl_tasks, 'r') as icl_f:
            icl_task_cfg = om.load(icl_f)
        icl_tasks_list = to_list_container(icl_task_cfg.icl_tasks)
    else:
        icl_tasks_list = icl_tasks

    def _validate_cfg(icl_cfg: dict[str, Any]):
        assert 'label' in icl_cfg
        assert 'dataset_uri' in icl_cfg and icl_cfg['dataset_uri'] is not None
        assert 'icl_task_type' in icl_cfg
        assert 'num_fewshot' in icl_cfg

        if 'metric_names' not in icl_cfg:
            if icl_cfg['icl_task_type'] == 'language_modeling':
                icl_cfg['metric_names'] = ['InContextLearningLMAccuracy']
            elif icl_cfg['icl_task_type'] == 'multiple_choice':
                icl_cfg['metric_names'] = [
                    'InContextLearningMultipleChoiceAccuracy',
                ]
            elif icl_cfg['icl_task_type'] == 'schema':
                icl_cfg['metric_names'] = [
                    'InContextLearningMultipleChoiceAccuracy',
                ]
            elif icl_cfg['icl_task_type'] == 'generation_task_with_answers':
                icl_cfg['metric_names'] = [
                    'InContextLearningGenerationExactMatchAccuracy',
                ]
            else:
                raise ValueError(
                    f'No metric_names defined, unable to build default metrics for icl_task_type={icl_cfg["icl_task_type"]}.',
                )

        if 'max_seq_len' not in icl_cfg:
            icl_cfg['max_seq_len'] = default_max_seq_len
        if 'batch_size' not in icl_cfg:
            icl_cfg['batch_size'] = default_batch_size

        if 'num_beams' in icl_cfg:
            raise ValueError(
                'num_beams is no longer supported as a top level icl_task parameter.'  + \
                'Please use generation_kwargs.num_beams instead.')

    for icl_cfg in icl_tasks_list:
        assert isinstance(
            icl_cfg,
            dict,
        ), f'Expected dict, got {type(icl_cfg)}, {icl_cfg=}'
        _validate_cfg(icl_cfg)
        for num_fewshot in list(icl_cfg['num_fewshot']):
            if tokenizer.pad_token_id is None:
                # Current workaround to support GPT2 tokenizer with `pad_token_id = None`
                pad_tok_id = tokenizer.eos_token_id
            else:
                pad_tok_id = tokenizer.pad_token_id

            label = f'{icl_cfg["label"]}/{num_fewshot}-shot'
            metric_names = list(icl_cfg['metric_names'])
            # TODO: fix Composer bug when copying local paths and destination exists
            destination_path = f'{destination_dir}/{icl_cfg["label"]}-{num_fewshot}.jsonl'
            if dist.get_local_rank() == 0 and os.path.exists(destination_path):
                os.remove(destination_path)
            dist.barrier()

            hf_parsing_map = icl_cfg.get('hf_parsing_map', {})
            hf_loading_vars = icl_cfg.get('hf_loading_vars', {})
            early_stopping_criteria = icl_cfg.get(
                'early_stopping_criteria',
                [],
            )
            # TODO: fix manual removal of non-constructor fields
            icl_constructor_kwargs = copy.deepcopy(icl_cfg)
            icl_constructor_kwargs.pop('label', None)
            icl_constructor_kwargs.pop('metric_names', None)
            icl_constructor_kwargs.pop('icl_task_type', None)
            icl_constructor_kwargs.pop('batch_size', None)
            icl_constructor_kwargs.pop('has_categories', None)

            # Add custom constructor arguments
            icl_constructor_kwargs['pad_tok_id'] = pad_tok_id
            icl_constructor_kwargs['num_fewshot'] = num_fewshot

            # Support backwards compatibility for the naming of "prelimiter" as "question_prelimiter"
            if 'question_prelimiter' in icl_constructor_kwargs:
                if 'prelimiter' in icl_constructor_kwargs:
                    raise ValueError(
                        'Both "question_prelimiter" and "prelimiter" are specified in the ICL task config. '
                        +
                        'Please only specify one of them, as they map to the same argument.',
                    )
                else:
                    icl_constructor_kwargs['prelimiter'
                                          ] = icl_constructor_kwargs.pop(
                                              'question_prelimiter',
                                          )

            assert early_stopping_criteria is None or isinstance(
                early_stopping_criteria,
                list,
            )

            dataloaders = get_icl_task_dataloader(
                icl_task_type=icl_cfg['icl_task_type'],
                dataset_uri=icl_cfg['dataset_uri'],
                tokenizer=tokenizer,
                batch_size=icl_cfg['batch_size'],
                hf_loading_vars=hf_loading_vars,
                hf_parsing_map=hf_parsing_map,
                has_categories=icl_cfg.get('has_categories', False),
                destination_path=destination_path,
                kwargs=icl_constructor_kwargs,
            )
            if 'has_categories' in icl_cfg and icl_cfg[
                'has_categories'] and isinstance(dataloaders, dict):
                for category in dataloaders.keys():
                    logger_keys.extend([
                        f'metrics/{label}/{category}/{m}' for m in metric_names
                    ])
                    evaluators.append(
                        Evaluator(
                            label=f'{label}/{category}',
                            dataloader=dataloaders[category],
                            metric_names=metric_names,
                        ),
                    )
            else:
                logger_keys.extend([
                    f'metrics/{label}/{m}' for m in metric_names
                ])
                evaluators.append(
                    Evaluator(
                        label=label,
                        dataloader=dataloaders,
                        metric_names=metric_names,
                        subset_num_batches=icl_subset_num_batches,
                    ),
                )

    return evaluators, logger_keys


@experimental_function('Tensor Parallelism')
def build_tp_strategies(
    name: str,
    model: ComposerModel,
) -> dict[str, ParallelStyle]:

    warnings.warn(
        'Checkpointing is not currently supported for tensor parallelism due to this pytorch bug: https://github.com/pytorch/pytorch/issues/134095#issuecomment-2345018244',
    )
    return construct_from_registry(
        name=name,
        registry=registry.tp_strategies,
        partial_function=False,
        kwargs={'model': model},
    )



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from llmfoundry.models.inference_api_wrapper.fmapi import (
    FMAPICasualLMEvalWrapper,
    FMAPIChatAPIEvalWrapper,
    FMAPIEvalInterface,
)
from llmfoundry.models.inference_api_wrapper.interface import \
    InferenceAPIEvalWrapper
from llmfoundry.models.inference_api_wrapper.openai_causal_lm import (
    OpenAICausalLMEvalWrapper,
    OpenAIChatAPIEvalWrapper,
    OpenAIEvalInterface,
)

__all__ = [
    'OpenAICausalLMEvalWrapper',
    'OpenAIChatAPIEvalWrapper',
    'OpenAIEvalInterface',
    'InferenceAPIEvalWrapper',
    'FMAPICasualLMEvalWrapper',
    'FMAPIChatAPIEvalWrapper',
    'FMAPIEvalInterface',
]



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from composer.metrics import (
    LanguageCrossEntropy,
    LanguagePerplexity,
    MaskedAccuracy,
)

from llmfoundry.eval.metrics import (
    InContextLearningGenerationExactMatchAccuracy,
    InContextLearningLMAccuracy,
    InContextLearningLMExpectedCalibrationError,
    InContextLearningMCExpectedCalibrationError,
    InContextLearningMultipleChoiceAccuracy,
)
from llmfoundry.metrics.token_acc import TokenAccuracy
from llmfoundry.registry import metrics

metrics.register('token_accuracy', func=TokenAccuracy)
metrics.register('lm_accuracy', func=InContextLearningLMAccuracy)
metrics.register(
    'lm_expected_calibration_error',
    func=InContextLearningLMExpectedCalibrationError,
)
metrics.register(
    'mc_expected_calibration_error',
    func=InContextLearningMCExpectedCalibrationError,
)
metrics.register('mc_accuracy', func=InContextLearningMultipleChoiceAccuracy)
metrics.register(
    'qa_accuracy',
    func=InContextLearningGenerationExactMatchAccuracy,
)
metrics.register('language_cross_entropy', func=LanguageCrossEntropy)
metrics.register('language_perplexity', func=LanguagePerplexity)
metrics.register('masked_accuracy', func=MaskedAccuracy)

DEFAULT_CAUSAL_LM_TRAIN_METRICS = [
    'language_cross_entropy',
    'language_perplexity',
    'token_accuracy',
]

DEFAULT_CAUSAL_LM_EVAL_METRICS = [
    'language_cross_entropy',
    'language_perplexity',
    'token_accuracy',
    'lm_accuracy',
    'lm_expected_calibration_error',
    'mc_expected_calibration_error',
    'mc_accuracy',
    'qa_accuracy',
]

DEFAULT_ENC_DEC_METRICS = [
    'language_cross_entropy',
    'masked_accuracy',
]

__all__ = [
    'TokenAccuracy',
    'InContextLearningLMAccuracy',
    'InContextLearningLMExpectedCalibrationError',
    'InContextLearningMCExpectedCalibrationError',
    'InContextLearningMultipleChoiceAccuracy',
    'InContextLearningGenerationExactMatchAccuracy',
    'DEFAULT_CAUSAL_LM_TRAIN_METRICS',
    'DEFAULT_CAUSAL_LM_EVAL_METRICS',
    'DEFAULT_ENC_DEC_METRICS',
]



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0
from llmfoundry.command_utils.data_prep.convert_dataset_hf import (
    convert_dataset_hf,
    convert_dataset_hf_from_args,
)
from llmfoundry.command_utils.data_prep.convert_dataset_json import (
    convert_dataset_json,
    convert_dataset_json_from_args,
)
from llmfoundry.command_utils.data_prep.convert_delta_to_json import (
    convert_delta_to_json_from_args,
    fetch_DT,
)
from llmfoundry.command_utils.data_prep.convert_finetuning_dataset import (
    convert_finetuning_dataset,
    convert_finetuning_dataset_from_args,
)
from llmfoundry.command_utils.data_prep.convert_text_to_mds import (
    convert_text_to_mds,
    convert_text_to_mds_from_args,
)
from llmfoundry.command_utils.eval import (
    eval_from_yaml,
    evaluate,
)
from llmfoundry.command_utils.train import (
    TRAIN_CONFIG_KEYS,
    TrainConfig,
    train,
    train_from_yaml,
    validate_config,
)

__all__ = [
    'train',
    'train_from_yaml',
    'TrainConfig',
    'TRAIN_CONFIG_KEYS',
    'validate_config',
    'evaluate',
    'eval_from_yaml',
    'convert_dataset_hf',
    'convert_dataset_hf_from_args',
    'convert_dataset_json',
    'convert_dataset_json_from_args',
    'convert_finetuning_dataset_from_args',
    'convert_finetuning_dataset',
    'convert_text_to_mds',
    'convert_text_to_mds_from_args',
    'convert_delta_to_json_from_args',
    'fetch_DT',
]



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import logging
import os
import re
import time
import urllib.parse
from collections import namedtuple
from concurrent.futures import ProcessPoolExecutor
from typing import TYPE_CHECKING, Iterable, Optional, Union
from uuid import uuid4

import google.protobuf.any_pb2 as any_pb2
import pandas as pd
import pyarrow as pa
import requests
from composer.utils import retry
from packaging import version

from llmfoundry.utils.exceptions import (
    ClusterDoesNotExistError,
    ClusterInvalidAccessMode,
    DeltaTableNotFoundError,
    FailedToConnectToDatabricksError,
    FailedToCreateSQLConnectionError,
    FaultyDataPrepCluster,
    InsufficientPermissionsError,
    UCNotEnabledError,
)

if TYPE_CHECKING:
    import pyspark.sql.connect.proto as pb2
    from databricks.sql.client import Connection as Connection
    from databricks.sql.client import Cursor as Cursor
    from pyspark.sql import SparkSession
    from pyspark.sql.connect.client.core import SparkConnectClient
    from pyspark.sql.connect.dataframe import DataFrame
    from pyspark.sql.dataframe import DataFrame as SparkDataFrame
    from pyspark.sql.types import Row

try:
    from pyspark.sql.connect.client.core import SparkConnectClient
    spark_connect_client_installed = True
except ImportError:
    spark_connect_client_installed = False

try:
    from pyspark.sql.connect.dataframe import DataFrame
    data_frame_installed = True
except ImportError:
    data_frame_installed = False

MINIMUM_DB_CONNECT_DBR_VERSION = '14.1'
MINIMUM_SQ_CONNECT_DBR_VERSION = '12.2'

TABLENAME_PATTERN = re.compile(r'(\S+)\.(\S+)\.(\S+)')

log = logging.getLogger(__name__)

Result = namedtuple(
    'Result',
    [
        'url',
        'row_count',
        'compressed_size',
        'uncompressed_size',
    ],
)  # pyright: ignore

# ``collect_as_cf`` is an addon new feature monkey patch on top of the DB Connect package.
# It allows the client to fetch the results in different formats from the server.
# To be able to use the code make sure this module is not overriden by DB Connect classes.


def to_cf(self: 'SparkConnectClient',
          plan: 'pb2.Plan',
          type: str = 'json') -> tuple[list[Result], int, bool]:
    """Executes the query plans and return as presigned URLS for cloud fetch.

    It can handle the current output formats that are supported by the server.
    In contrast to the regular API methods of the client, this method does not
    return the schema and drops all other responses.

    Args:
        self (SparkConnectClient): The SparkConnectClient we are processing.
        plan (pb2.Plan): The plan object to be executed by spark.
        type (str): The output format of the result, supported formats are 'json', 'csv', and 'arrow'.

    Returns:
        Tuple[List[Result], int, bool]: A tuple containing:
            - A list of Result namedtuples, each containing a URL, row count, compressed size,
                and uncompressed size of the part of the result.
            - Total row count of all parts of the result.
            - A boolean indicating whether the result has been truncated.
    """
    req = self._execute_plan_request_with_metadata()
    req.plan.CopyFrom(plan)

    import pyspark.sql.connect.proto as pb2
    import pyspark.sql.connect.proto.cloud_pb2 as cloud_pb2

    # Add the request options
    if type == 'json':
        format = cloud_pb2.ResultOptions.CloudOptions.FORMAT_JSON
    elif type == 'csv':
        format = cloud_pb2.ResultOptions.CloudOptions.FORMAT_CSV
    elif type == 'arrow':
        format = cloud_pb2.ResultOptions.CloudOptions.FORMAT_ARROW
    else:
        raise ValueError(
            f'Only formats json, csv, and arrow are supported. Got invalid type {type}',
        )

    ro = cloud_pb2.ResultOptions(
        type=cloud_pb2.ResultOptions.TYPE_CLOUD,
        cloudOptions=cloud_pb2.ResultOptions.CloudOptions(
            format=format,
            useCompression=False,
        ),
    )
    cloud_option = any_pb2.Any()
    cloud_option.Pack(ro)
    req.request_options.append(
        pb2.ExecutePlanRequest.RequestOption(extension=cloud_option),
    )

    # Create the iterator
    from pyspark.sql.connect.client.reattach import (
        ExecutePlanResponseReattachableIterator,
    )
    iterator = ExecutePlanResponseReattachableIterator(
        req,
        self._stub,
        self._retry_policy,
        self._builder.metadata(),
    )
    # Iterate over the response
    result = []
    row_count = 0
    is_overflow = False

    for response in iterator:
        if response.HasField('extension') and response.extension.Is(
            cloud_pb2.CloudResultBatch.DESCRIPTOR,
        ):
            batch = cloud_pb2.CloudResultBatch()
            if not response.extension.Is(cloud_pb2.CloudResultBatch.DESCRIPTOR):
                raise ValueError(
                    'Response extension is not of type CloudResultBatch.',
                )
            response.extension.Unpack(batch)
            result += [
                Result(
                    b.url,
                    b.row_count,
                    b.compressed_size,
                    b.uncompressed_size,
                ) for b in batch.results
            ]
            row_count += sum(result.row_count for result in batch.results)
            is_overflow |= batch.truncated
    return result, row_count, is_overflow


if spark_connect_client_installed:
    SparkConnectClient.to_cf = to_cf  # pyright: ignore


def collect_as_cf(self: 'DataFrame',
                  type: str = 'json') -> tuple[list[Result], int, bool]:
    """Collects DataFrame execution plan as presigned URLs.

    This method is a wrapper around the `to_cf` method of SparkConnectClient. It takes the
    execution plan of the current DataFrame, converts it to a protocol buffer format, and then
    uses the `to_cf` method to execute the plan and fetch results as presigned URLs.

    Args:
        self (pd.DataFrame): The dataframe we are processing.
        type (str): The output format of the result, supported formats are 'json', 'csv', and 'arrow'.

    Returns:
        Tuple[List[Result], int, bool]: A tuple containing:
            - A list of Result namedtuples, each containing a URL, row count, compressed size,
              and uncompressed size of the part of the result.
            - Total row count of all parts of the result.
            - A boolean indicating whether the result is truncated or overflowed.
    """
    query = self._plan.to_proto(self._session.client)  # pyright: ignore
    return self._session.client.to_cf(query, type)  # pyright: ignore


if data_frame_installed:
    DataFrame.collect_cf = collect_as_cf  # pyright: ignore


def iterative_combine_jsons(json_directory: str, output_file: str) -> None:
    """Combine jsonl files in json_directory into one big jsonl file.

    This function does not work for nested subdirectories.

    Args:
        json_directory(str): directory containing the JSONL files
        output_file(str): path to the output combined JSONL file
    """
    json_files = [f for f in os.listdir(json_directory) if f.endswith('.jsonl')]
    with open(output_file, 'w') as outfile:
        for file_name in json_files:
            with open(os.path.join(json_directory, file_name), 'r') as infile:
                for line in infile:
                    outfile.write(line)
    log.info('JSON files have been combined into a JSONL file.')


def run_query(
    query: str,
    method: str,
    cursor: Optional['Cursor'] = None,
    spark: Optional['SparkSession'] = None,
    collect: bool = True,
) -> Optional[Union[list['Row'], 'DataFrame', 'SparkDataFrame']]:
    """Run SQL query via databricks-connect or databricks-sql.

    Args:
        query (str): sql query
        method (str): select from dbsql and dbconnect
        cursor (Optional[Cursor]): connection.cursor
        spark (Optional[SparkSession]): spark session
        collect (bool): whether to get the underlying data from spark dataframe
    """
    if method == 'dbsql':
        if cursor is None:
            raise ValueError(f'cursor cannot be None if using method dbsql')
        cursor.execute(query)
        if collect:
            return cursor.fetchall()
    elif method == 'dbconnect':
        if spark == None:
            raise ValueError(f'sparkSession is required for dbconnect')
        df = spark.sql(query)
        if collect:
            return df.collect()
        return df
    else:
        raise ValueError(f'Unrecognized method: {method}')


def get_args(signed: list, json_output_folder: str, columns: list) -> Iterable:
    for i, r in enumerate(signed):
        yield (i, r.url, json_output_folder, columns)


def download(
    ipart: int,
    url: str,
    json_output_folder: str,
    columns: Optional[list] = None,
    resp_format: str = 'arrow',
    compressed: bool = False,
) -> None:
    """Thread download presigned url and save to jsonl locally.

    Args:
        ipart (int): presigned url id
        url (str): presigned url
        json_output_folder (str): directory to save the ipart_th segment of dataframe
        columns (list): schema to save to json
        resp_format (str): whether to use arrow or json when collect
        compressed (bool): if data is compressed before downloading. Need decompress if compressed=True.
    """
    resp = requests.get(url)
    if resp.status_code == 200:
        if resp_format == 'json':
            data = resp.json()
            pd.DataFrame(data, columns=columns).to_json(
                os.path.join(
                    json_output_folder,
                    'part_' + str(ipart) + '.jsonl',
                ),
                orient='records',
                lines=True,
            )
            return

        # When resp_format is arrow:
        if compressed:
            # The data is lz4 compressed arrow format.
            # Decompress the data
            import lz4.frame
            decompressed_data = lz4.frame.decompress(resp.content)
            # Convert the decompressed data into a PyArrow table
            reader = pa.ipc.open_stream(decompressed_data)
        else:
            reader = pa.ipc.open_stream(resp.content)
        table = reader.read_all()

        # Convert the PyArrow table into a pandas DataFrame
        df = table.to_pandas()
        df.to_json(
            os.path.join(json_output_folder, 'part_' + str(ipart) + '.jsonl'),
            orient='records',
            lines=True,
            force_ascii=False,
        )


def download_starargs(args: tuple) -> None:
    return download(*args)


def format_tablename(table_name: str) -> str:
    """Escape catalog, schema and table names with backticks.

    This needs to be done when running SQL queries/setting spark sessions to prevent invalid identifier errors.

    Args:
        table_name (str): catalog.scheme.tablename on UC
    """
    match = re.match(TABLENAME_PATTERN, table_name)

    if match is None:
        return table_name

    formatted_identifiers = []
    for i in range(1, 4):
        identifier = f'`{match.group(i)}`'
        formatted_identifiers.append(identifier)

    return '.'.join(formatted_identifiers)


def fetch_data(
    method: str,
    cursor: Optional['Cursor'],
    sparkSession: Optional['SparkSession'],
    start: int,
    end: int,
    order_by: str,
    tablename: str,
    columns_str: str,
    json_output_folder: str,
) -> None:
    """Fetches a specified range of rows from a given table to a json file.

    This function executes a SQL query to retrieve a range of rows, determined by 'start' and 'end' indexes,
    from a specified table and column set. The fetched data is then exported as a JSON file.

    Args:
        method (str): The method to use for fetching data, either 'dbconnect' or 'dbsql'.
        cursor (Optional[Cursor]): The cursor object for executing queries in 'dbsql' method.
        sparkSession (Optional[SparkSession]): The Spark session object for executing queries in 'dbconnect' method.
        start (int): The starting index for row fetching.
        end (int): The ending index for row fetching.
        order_by (str): The column name to use for ordering the rows.
        tablename (str): The name of the table from which to fetch the data.
        columns_str (str): The string representation of the columns to select from the table.
        json_output_folder (str): The file path where the resulting JSON file will be saved.

    Returns:
        None: The function doesn't return any value, but writes the result to a JSONL file.
    """
    query = f"""
    WITH NumberedRows AS (
        SELECT
            *,
            ROW_NUMBER() OVER (ORDER BY {order_by}) AS rn
        FROM
            {tablename}
    )
    SELECT {columns_str}
    FROM NumberedRows
    WHERE rn BETWEEN {start+1} AND {end}"""

    if method == 'dbconnect':
        spark_df = run_query(query, method, cursor, sparkSession, collect=False)
        if spark_df is None:
            raise RuntimeError(
                f'Expect spark dataframe with {query} but got None',
            )
        pdf = spark_df.toPandas()  # pyright: ignore
    else:  #  method == 'dbsql':
        ans = run_query(query, method, cursor, sparkSession, collect=True)
        if ans is None:
            raise RuntimeError(f'Got empty results with {query}')
        records = [r.asDict() for r in ans]  # pyright: ignore
        pdf = pd.DataFrame.from_dict(records)

    pdf.to_json(
        os.path.join(json_output_folder, f'part_{start+1}_{end}.jsonl'),
        orient='records',
        lines=True,
    )


@retry(Exception, num_attempts=5, initial_backoff=1.0, max_jitter=0.5)
def get_total_rows(
    tablename: str,
    method: str,
    cursor: Optional['Cursor'],
    sparkSession: Optional['SparkSession'],
):
    ans = run_query(
        f'SELECT COUNT(*) FROM {tablename}',
        method,
        cursor,
        sparkSession,
    )
    nrows = [row.asDict() for row in ans][0].popitem()[1]  # pyright: ignore
    log.info(f'total_rows = {nrows}')
    return nrows


@retry(Exception, num_attempts=5, initial_backoff=1.0, max_jitter=0.5)
def get_columns_info(
    tablename: str,
    method: str,
    cursor: Optional['Cursor'],
    sparkSession: Optional['SparkSession'],
):
    ans = run_query(
        f'SHOW COLUMNS IN {tablename}',
        method,
        cursor,
        sparkSession,
    )
    columns = [row.asDict().popitem()[1] for row in ans]  # pyright: ignore
    order_by = columns[0]
    columns_str = ','.join(columns)
    log.info(f'order by column {order_by}')
    return columns, order_by, columns_str


def fetch(
    method: str,
    tablename: str,
    json_output_folder: str,
    batch_size: int = 1 << 30,
    processes: int = 1,
    sparkSession: Optional['SparkSession'] = None,
    dbsql: Optional['Connection'] = None,
) -> None:
    """Fetch UC delta table with databricks-connect as JSONL.

    Args:
        method (str): dbconnect or dbsql
        tablename (str): catalog.scheme.tablename on UC
        json_output_folder (str): path to write the result json file to
        batch_size (int): number of rows that dbsql fetches each time to avoid OOM
        processes (int): max number of processes to use to parallelize the fetch
        sparkSession (pyspark.sql.sparksession): spark session
        dbsql (databricks.sql.connect): dbsql session
    """
    cursor = dbsql.cursor() if dbsql is not None else None
    try:
        # Get total rows
        nrows = get_total_rows(tablename, method, cursor, sparkSession)

        # Get columns info
        columns, order_by, columns_str = get_columns_info(
            tablename,
            method,
            cursor,
            sparkSession,
        )

        if method == 'dbconnect' and sparkSession is not None:
            log.info(f'{processes=}')
            df = sparkSession.table(tablename)

            # Running the query and collecting the data as arrow or json.
            signed, _, _ = df.collect_cf('arrow')  # pyright: ignore
            log.info(f'len(signed) = {len(signed)}')

            args = get_args(signed, json_output_folder, columns)

            # Stopping the SparkSession to avoid spilling connection state into the subprocesses.
            sparkSession.stop()

            with ProcessPoolExecutor(max_workers=processes) as executor:
                list(executor.map(download_starargs, args))

        elif method == 'dbsql' and cursor is not None:
            for start in range(0, nrows, batch_size):
                log.warning(f'batch {start}')
                end = min(start + batch_size, nrows)
                fetch_data(
                    method,
                    cursor,
                    sparkSession,
                    start,
                    end,
                    order_by,
                    tablename,
                    columns_str,
                    json_output_folder,
                )

    except Exception as e:
        from databricks.sql.exc import ServerOperationError
        from pyspark.errors import AnalysisException

        if isinstance(e, (AnalysisException, ServerOperationError)):
            if 'INSUFFICIENT_PERMISSIONS' in str(e):
                raise InsufficientPermissionsError(str(e)) from e
            elif 'UC_NOT_ENABLED' in str(e):
                raise UCNotEnabledError() from e
            elif 'DELTA_TABLE_NOT_FOUND' in str(e):
                err_str = str(e)
                # Error string should be in this format:
                # ---
                # Error processing `catalog`.`volume_name`.`table_name`:
                # [DELTA_TABLE_NOT_FOUND] Delta table `volume_name`.`table_name`
                # doesn't exist.
                # ---
                parts = err_str.split('`')
                if len(parts) < 7:
                    # Failed to parse error, our codebase is brittle
                    # with respect to the string representations of
                    # errors in the spark library.
                    catalog_name, volume_name, table_name = ['unknown'] * 3
                else:
                    catalog_name = parts[1]
                    volume_name = parts[3]
                    table_name = parts[5]
                raise DeltaTableNotFoundError(
                    catalog_name,
                    volume_name,
                    table_name,
                ) from e

        if isinstance(e, InsufficientPermissionsError):
            raise

        # For any other exception, raise a general error
        raise RuntimeError(f'Error processing {tablename}: {str(e)}') from e

    finally:
        if cursor is not None:
            cursor.close()


def validate_and_get_cluster_info(
    cluster_id: Optional[str],
    databricks_host: str,
    databricks_token: str,
    http_path: Optional[str],
    use_serverless: bool = False,
) -> tuple:
    """Validate and get cluster info for running the Delta to JSONL conversion.

    Args:
        cluster_id (str): cluster id to validate and fetch additional info for
        databricks_host (str): databricks host name
        databricks_token (str): databricks auth token
        http_path (Optional[str]): http path to use for sql connect
        use_serverless (bool): whether to use serverless or not
    """
    method = 'dbsql'
    dbsql = None
    sparkSession = None

    if use_serverless:
        method = 'dbconnect'
    else:
        if not cluster_id:
            raise ValueError(
                'cluster_id is not set, however use_serverless is False',
            )
        from databricks.sdk import WorkspaceClient
        w = WorkspaceClient()
        res = w.clusters.get(cluster_id=cluster_id)
        if res is None:
            raise ClusterDoesNotExistError(cluster_id)

        data_security_mode = str(
            res.data_security_mode,
        ).upper()[len('DATASECURITYMODE.'):]

        # NONE stands for No Isolation Shared
        # This check actually checks for Unity Catalog governance compatibility and does not
        # check for invalid cluster access for a particular user. Cluster access controls is
        # difficult and there is no single existing API to check this.
        if data_security_mode == 'NONE':
            raise ClusterInvalidAccessMode(
                cluster_id=cluster_id,
                access_mode=data_security_mode,
            )

        assert res.spark_version is not None
        stripped_runtime = re.sub(
            r'[a-zA-Z]',
            '',
            res.spark_version.split('-scala')
            [0].replace(  # type: ignore
                'x-snapshot', '',
            ),
        )
        runtime_version = re.sub(r'[.-]*$', '', stripped_runtime)
        if version.parse(
            runtime_version,
        ) < version.parse(MINIMUM_SQ_CONNECT_DBR_VERSION):
            raise ValueError(
                f'The minium DBR version required is {MINIMUM_SQ_CONNECT_DBR_VERSION} but got {version.parse(runtime_version)}',
            )

        if http_path is None and version.parse(
            runtime_version,
        ) >= version.parse(MINIMUM_DB_CONNECT_DBR_VERSION):
            method = 'dbconnect'

    if method == 'dbconnect':
        from databricks.connect import DatabricksSession
        try:
            if use_serverless:
                session_id = str(uuid4())
                sparkSession = DatabricksSession.builder.host(
                    databricks_host,
                ).token(
                    databricks_token,
                ).header('x-databricks-session-id', session_id).getOrCreate()

            else:
                if not cluster_id:
                    raise ValueError('cluster_id is needed for dbconnect.',)
                sparkSession = DatabricksSession.builder.remote(
                    host=databricks_host,
                    token=databricks_token,
                    cluster_id=cluster_id,
                ).getOrCreate()

        except Exception as e:
            raise FailedToConnectToDatabricksError() from e
    else:
        try:
            from databricks import sql
            dbsql = sql.connect(
                server_hostname=re.compile(r'^https?://').sub(
                    '', databricks_host).strip(
                    ),  # sqlconnect hangs if hostname starts with https
                http_path=http_path,
                access_token=databricks_token,
            )
        except Exception as e:
            raise FailedToCreateSQLConnectionError() from e
    return method, dbsql, sparkSession


def fetch_DT(
    delta_table_name: str,
    json_output_folder: str,
    http_path: Optional[str],
    cluster_id: Optional[str],
    use_serverless: bool,
    DATABRICKS_HOST: str,
    DATABRICKS_TOKEN: str,
    batch_size: int = 1 << 30,
    processes: int = os.cpu_count(), # type: ignore
    json_output_filename: str = 'train-00000-of-00001.jsonl',
) -> None:
    """Fetch UC Delta Table to local as jsonl."""
    log.info(f'Start .... Convert delta to json')

    obj = urllib.parse.urlparse(json_output_folder)
    if obj.scheme != '':
        raise ValueError(
            'Check the json_output_folder and verify it is a local path!',
        )

    if os.path.exists(json_output_folder):
        if not os.path.isdir(json_output_folder) or os.listdir(
            json_output_folder,
        ):
            raise RuntimeError(
                f'Output folder {json_output_folder} already exists and is not empty. Please remove it and retry.',
            )

    os.makedirs(json_output_folder, exist_ok=True)

    if not json_output_filename.endswith('.jsonl'):
        raise ValueError('json_output_filename needs to be a jsonl file')

    log.info(f'Directory {json_output_folder} created.')

    # validate_and_get_cluster_info allows cluster_id to be None if use_serverless is True
    method, dbsql, sparkSession = validate_and_get_cluster_info(
        cluster_id=cluster_id,
        databricks_host=DATABRICKS_HOST,
        databricks_token=DATABRICKS_TOKEN,
        http_path=http_path,
        use_serverless=use_serverless,
    )

    formatted_delta_table_name = format_tablename(delta_table_name)
    import grpc
    import pyspark.errors.exceptions.connect as spark_errors
    try:
        fetch(
            method,
            formatted_delta_table_name,
            json_output_folder,
            batch_size,
            processes,
            sparkSession,
            dbsql,
        )
    except (grpc.RpcError, spark_errors.SparkConnectGrpcException) as e:
        if isinstance(
            e,
            spark_errors.SparkConnectGrpcException,
        ) and 'Cannot start cluster' in str(e):
            raise FaultyDataPrepCluster(
                message=
                f'The data preparation cluster you provided is terminated. Please retry with a cluster that is healthy and alive. {e}',
            ) from e
        if isinstance(
            e,
            spark_errors.SparkConnectGrpcException,
        ) and 'is not usable' in str(e):
            raise FaultyDataPrepCluster(
                message=
                f'The data preparation cluster you provided is not usable. Please retry with a cluster that is healthy and alive. {e}',
            ) from e
        if isinstance(e, grpc.RpcError) and e.code(
        ) == grpc.StatusCode.INTERNAL and 'Job aborted due to stage failure' in e.details(
        ):
            raise FaultyDataPrepCluster(
                message=
                f'Faulty data prep cluster, please try swapping data prep cluster: {e.details()}',
            ) from e
        raise e

    if dbsql is not None:
        dbsql.close()

    # combine downloaded jsonl into one big jsonl for IFT
    iterative_combine_jsons(
        json_output_folder,
        os.path.join(json_output_folder, json_output_filename),
    )


def _check_imports():
    try:
        import lz4.frame
        _ = lz4.frame
    except ImportError as e:
        raise ImportError('lz4 is not installed.') from e

    try:
        from databricks.connect import DatabricksSession
        _ = DatabricksSession
    except ImportError as e:
        raise ImportError(
            'databricks-connect is not installed or improperly configured.',
        ) from e

    try:
        from databricks import sql
        from databricks.sdk import WorkspaceClient
        from databricks.sql.client import Connection as Connection
        from databricks.sql.client import Cursor as Cursor
        _ = WorkspaceClient, Connection, Cursor, sql
    except ImportError as e:
        raise ImportError(
            'databricks-sdk is not installed or improperly configured.',
        ) from e

    try:
        import pyspark.sql.connect.proto as pb2
        import pyspark.sql.connect.proto.cloud_pb2 as cloud_pb2
        from pyspark.sql import SparkSession
        from pyspark.sql.connect.client.core import SparkConnectClient
        from pyspark.sql.connect.client.reattach import (
            ExecutePlanResponseReattachableIterator,
        )
        from pyspark.sql.connect.dataframe import DataFrame
        from pyspark.sql.dataframe import DataFrame as SparkDataFrame
        from pyspark.sql.types import Row
        _ = (
            pb2,
            cloud_pb2,
            SparkSession,
            SparkConnectClient,
            ExecutePlanResponseReattachableIterator,
            DataFrame,
            SparkDataFrame,
            Row,
        )
    except ImportError as e:
        raise ImportError(
            'pyspark is not installed or improperly configured.',
        ) from e


def convert_delta_to_json_from_args(
    delta_table_name: str,
    json_output_folder: str,
    http_path: Optional[str],
    cluster_id: Optional[str],
    use_serverless: bool,
    batch_size: int,
    processes: int,
    json_output_filename: str,
) -> None:
    """A wrapper for `convert_dataset_json` that parses arguments.

    Args:
        delta_table_name (str): UC table <catalog>.<schema>.<table name>
        json_output_folder (str): Local path to save the converted json
        http_path (Optional[str]): If set, dbsql method is used
        batch_size (int): Row chunks to transmit a time to avoid OOM
        processes (int): Number of processes allowed to use
        cluster_id (Optional[str]): Cluster ID with runtime newer than 14.1.0 and access mode of either assigned or shared can use databricks-connect.
        use_serverless (bool): Use serverless or not. Make sure the workspace is entitled with serverless
        json_output_filename (str): The name of the combined final jsonl that combines all partitioned jsonl
    """
    os.environ['WORLD_SIZE'] = '1'
    _check_imports()
    from databricks.sdk import WorkspaceClient
    w = WorkspaceClient()
    DATABRICKS_HOST = w.config.host
    DATABRICKS_TOKEN = w.config.token

    tik = time.time()
    fetch_DT(
        delta_table_name=delta_table_name,
        json_output_folder=json_output_folder,
        http_path=http_path,
        batch_size=batch_size,
        processes=processes,
        cluster_id=cluster_id,
        use_serverless=use_serverless,
        json_output_filename=json_output_filename,
        DATABRICKS_HOST=DATABRICKS_HOST,
        DATABRICKS_TOKEN=DATABRICKS_TOKEN,
    )
    log.info(f'Elapsed time {time.time() - tik}')



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import logging

from composer.core import Callback, State
from composer.loggers import Logger

from llmfoundry.utils.warnings import experimental_class

__all__ = [
    'GlobalLRScaling',
    'LayerFreezing',
]

log = logging.getLogger(__name__)


@experimental_class('GlobalLRScaling')
class GlobalLRScaling(Callback):
    """GlobalLRScaling.

    This callback can be applied upon resuming a model checkpoint. Upon
    fit_start it will multiply the base LR by `lr_scale` and set the WD to be.

    `wd_pct` * `lr`.

    Args:
        lr_scale (float): Multiplicative factor to scale LR by
        wd_pct (float): Percentage of LR to set weight decay to.
    """

    def __init__(self, lr_scale: float, wd_pct: float = 0.0):
        self.lr_scale = lr_scale
        self.wd_pct = wd_pct

    def fit_start(self, state: State, logger: Logger) -> None:
        del logger  # unused

        if hasattr(state, 'optimizer') and state.optimizers is None:
            raise Exception('No optimizers defined')
        for optimizer in state.optimizers:
            for group in optimizer.param_groups:
                group['lr'] *= self.lr_scale
                group['weight_decay'] = group['lr'] * self.wd_pct
                if 'initial_lr' in group:
                    group['initial_lr'] *= self.lr_scale
                log.info(
                    f"Set LR and WD to {group['lr']}, {group['weight_decay']}",
                )

        for scheduler in state.schedulers:
            scheduler.base_lrs = [
                self.lr_scale * lr for lr in scheduler.base_lrs
            ]


@experimental_class('LayerFreezing')
class LayerFreezing(Callback):
    """LayerFreezing.

    This callback can be applied upon resuming a model checkpoint. Upon
    fit_start it freeze the layers specified in `layer_names`. If using
    activation checkpointing, please set the
    `activation_checkpointing_reentrant` flag in `fsdp_config` to false.

    Args:
        layer_names (float): Names of layers to freeze.
    """

    def __init__(self, layer_names: list[str]):
        self.layer_names = set(layer_names)

    def fit_start(self, state: State, logger: Logger) -> None:
        del logger  # unused

        model_layers = {name for name, _ in state.model.named_parameters()}
        for layer in self.layer_names:
            if layer not in model_layers:
                raise Exception(
                    f'Attempted to freeze layer not found in model: {layer}\nAvailable layers: {model_layers}',
                )

        successful_freeze = False
        for name, p in state.model.named_parameters():
            if p.requires_grad and name in self.layer_names:
                p.requires_grad = False
                log.debug(f'Froze layer: {name}\nParam: {p}')
                successful_freeze = True

        if not successful_freeze:
            raise Exception(
                f"Tried to run LayerFreezing but didn't freeze any layers",
            )



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import tempfile

from composer.utils import dist


def dist_mkdtemp() -> str:
    """Creates a temp directory on local rank 0 to use for other ranks.

    Returns:
        str: The path to the temporary directory.
    """
    tempdir = None
    local_rank = dist.get_local_rank()
    global_rank = dist.get_global_rank()
    if local_rank == 0:
        tempdir = tempfile.mkdtemp()

    tempdir = dist.all_gather_object(tempdir)[global_rank - local_rank]
    if tempdir is None:
        raise RuntimeError('Dist operation to get tempdir failed.')
    return tempdir



# Copyright 2024 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

from typing import Any

import torch

from llmfoundry.layers_registry import (
    attention_classes,
    ffns,
    ffns_with_megablocks,
    ffns_with_norm,
    norms,
)
from llmfoundry.models.layers.blocks import FusedNormAttentionNorm, MPTBlock

__all__ = [
    'build_act_ckpt_mod_to_blocks',
    'pass_on_block_idx',
    'check_mapping_blocks_overlap',
]


def pass_on_block_idx(parent: torch.nn.Module):
    if not hasattr(parent, 'block_idx') or not hasattr(parent, 'max_block_idx'):
        return
    for child in parent.children():
        child.block_idx = parent.block_idx
        child.max_block_idx = parent.max_block_idx
        if child.children():
            pass_on_block_idx(child)


def get_act_ckpt_module(mod_name: str) -> Any:
    """Get the module type from the module name."""
    if mod_name.lower() == 'mptblock':
        mod_type = MPTBlock
    elif mod_name in attention_classes:
        mod_type = attention_classes.get(mod_name)
    elif mod_name.lower() == 'norm_attn_norm':
        mod_type = FusedNormAttentionNorm
    elif mod_name in ffns:
        mod_type = ffns.get(mod_name)
    elif mod_name in ffns_with_norm:
        mod_type = ffns_with_norm.get(mod_name)
    elif mod_name in ffns_with_megablocks:
        mod_type = ffns_with_megablocks.get(mod_name)
    elif mod_name in norms:
        mod_type = norms.get(mod_name)
    else:
        msg = ', '.join(
            list(attention_classes.get_all()) + list(ffns.get_all()) +
            list(ffns_with_norm.get_all()) +
            list(ffns_with_megablocks.get_all()) + list(norms.get_all()) +
            ['MPTBlock'],
        )
        raise ValueError(
            f'{mod_name} (specified in activation_checkpointing_target) is not a recognized option out of available options {msg}.',
        )
    return mod_type


def parse_ele_str(ele: str, max_block_idx: int) -> list:
    """Parse a string in target_blocks and return a list of block ids to add.

    Supported formats are: first-n, middle-m, last-k, range-i-j which correspond
    to the first n, the middle m,  the last k, and the range [i, j).
    """
    to_add = None
    if ele.startswith('first-'):
        assert ele[6:].isdigit(), f'Invalid target_blocks element {ele}'
        to_add = list(range(min(int(ele[6:]), max_block_idx + 1)))
    elif ele.startswith('last-'):
        assert ele[5:].isdigit(), f'Invalid target_blocks element {ele}'
        to_add = list(
            range(max(max_block_idx - int(ele[5:]) + 1, 0), max_block_idx + 1),
        )
    elif ele.startswith('middle-'):
        assert ele[7:].isdigit(), f'Invalid target_blocks element {ele}'
        num = int(ele[7:])
        start = max(max_block_idx // 2 - num // 2, 0)
        end = min(start + num, max_block_idx + 1)
        to_add = list(range(start, end))
    elif ele.startswith('range-'):
        r = ele[6:].split('-')
        assert len(r) == 2, f'Invalid target_blocks element {ele}'
        start, end = int(r[0]), int(r[1])
        start = max(start, 0)
        end = min(end, max_block_idx + 1)
        to_add = list(range(start, end))
    else:
        raise ValueError(f'Invalid target_blocks element {ele}')
    return to_add


def get_target_block_list(target_blocks: Any, max_block_idx: int) -> list:
    """Parse the user input and return a list of block ids."""
    candidate_block_ids = []
    if isinstance(target_blocks, int):
        candidate_block_ids = list(range(target_blocks))
    elif isinstance(target_blocks, list):
        for ele in target_blocks:
            if isinstance(ele, int):
                candidate_block_ids.append(ele)
            elif isinstance(ele, str):
                to_add = parse_ele_str(ele, max_block_idx)
                candidate_block_ids.extend(to_add)
            else:
                raise ValueError(
                    f'target_blocks must be a list of integers or "first-n", "middle-m", "last-k", or "range-i-j" where n, m, k, i, j are integers, but got {target_blocks}',
                )
    elif isinstance(target_blocks, str):
        target_blocks = target_blocks.replace(' ', '')
        for ele in target_blocks.split(','):
            to_add = parse_ele_str(ele, max_block_idx)
            candidate_block_ids.extend(to_add)
    else:
        raise ValueError(
            f'target_blocks must be either a single integer, or a list of integers, or a comma separated string made of "first-n", "last-m", "middle-k", "range-i-j", or a list of mixed integers and before-mentioned strings, but got {type(target_blocks)}',
        )

    candidate_block_ids = list(set(candidate_block_ids))
    return candidate_block_ids


def check_mapping_blocks_overlap(mapping: dict, max_block_idx: int) -> None:
    """Check if the block ids in the mapping overlap with each other."""
    all_blocks = [None] * (max_block_idx + 1)
    for k, v in mapping.items():
        if v == -1:
            v = list(range(max_block_idx + 1))
        for vv in v:
            if vv < 0 or vv > max_block_idx:
                continue
            else:
                if all_blocks[vv] is not None:
                    raise ValueError(
                        f'Block {vv} is assigned to both {k} and {all_blocks[vv]}. Each block can only have one granularity of activation checkpointing. Make sure the target_blocks in activation_checkpointing_target do not overlap. For more details, refer to the docs of activation_checkpointing_fn.',
                    )
                else:
                    all_blocks[vv] = k


def build_act_ckpt_mod_to_blocks(
    act_ckpt_target: Any,
    top_module: Any,
    max_block_idx: int,
) -> dict:
    act_ckpt_mod_to_blocks = {}
    if act_ckpt_target is None or act_ckpt_target == []:
        mod = top_module
        act_ckpt_mod_to_blocks[mod] = -1
    elif isinstance(act_ckpt_target, str):
        mod = get_act_ckpt_module(act_ckpt_target)
        act_ckpt_mod_to_blocks[mod] = -1
    elif isinstance(act_ckpt_target, list):
        for target in act_ckpt_target:
            mod = get_act_ckpt_module(target)
            act_ckpt_mod_to_blocks[mod] = -1
    elif isinstance(act_ckpt_target, dict):
        for k, v in act_ckpt_target.items():
            mod = get_act_ckpt_module(k)
            block_ids = get_target_block_list(v, max_block_idx)
            act_ckpt_mod_to_blocks[mod] = block_ids
    else:
        raise ValueError(
            f'activation_checkpointing_target must be either a single string or a list or a dict, but got {type(act_ckpt_target)}',
        )

    return act_ckpt_mod_to_blocks



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

"""Datasets for converting to MDS Shards."""
import os
import warnings
from abc import ABC, abstractmethod
from typing import Iterable, Optional, Union

import datasets as hf_datasets
import numpy as np
from numpy.typing import NDArray
from torch.utils.data import IterableDataset
from transformers import PreTrainedTokenizerBase

__all__ = [
    'AbstractConcatTokensDataset',
    'ConcatTokensDataset',
    'NoConcatDataset',
    'stream_remote_local_validate',
    'SUPPORTED_MDS_ENCODING_TYPES',
]

SUPPORTED_MDS_ENCODING_TYPES = [
    'int8',
    'int16',
    'int32',
    'int64',
    'uint8',
    'uint16',
    'uint32',
    'uint64',
]


class NoConcatDataset(IterableDataset):
    """An IterableDataset that returns text samples for MDSWriter.

    Returns dicts of {'text': bytes}
    """

    def __init__(
        self,
        hf_dataset: Union[hf_datasets.IterableDataset, hf_datasets.Dataset],
    ):
        self.hf_dataset = hf_dataset

    def __iter__(self) -> Iterable[dict[str, bytes]]:
        for sample in self.hf_dataset:
            # convert to bytes to store in MDS binary format
            yield {'text': sample['text'].encode('utf-8')}


class AbstractConcatTokensDataset(ABC, IterableDataset):
    """Abstract class for defining an IterableDataset that tokenizes and.

    concatenates text samples on the fly.
    """

    def __init__(
        self,
        tokenizer: PreTrainedTokenizerBase,
        max_length: int,
        bos_text: str,
        eos_text: str,
        no_wrap: bool,
    ):
        self.tokenizer = tokenizer
        os.environ['TOKENIZERS_PARALLELISM'] = 'false'
        self.max_length = max_length
        self.bos_text = bos_text
        self.eos_text = eos_text
        self.should_wrap = not no_wrap

        self.bos_tokens = self.tokenizer(
            self.bos_text,
            truncation=False,
            padding=False,
            add_special_tokens=False,
        )['input_ids']
        if len(self.bos_tokens) > 1:
            warnings.warn(
                f'You specified --concat_tokens with --bos_text, but your BOS text is not tokenizing to one token\
                , instead we got {self.bos_tokens}. Quit if this was in error.',
            )

        self.eos_tokens = self.tokenizer(
            self.eos_text,
            truncation=False,
            padding=False,
            add_special_tokens=False,
        )['input_ids']
        if len(self.eos_tokens) > 1:
            warnings.warn(
                f'You specified --concat_tokens with --eos_text, but your EOS text is not tokenizing to one token\
                , instead we got {self.eos_tokens}. Quit if this was in error.',
            )

        eos_text_provided = self.eos_text != ''
        bos_text_provided = self.bos_text != ''
        test_text = self.tokenizer('')
        if len(
            test_text['input_ids'],
        ) > 0 and (eos_text_provided or bos_text_provided):
            message = 'both eos and bos' if eos_text_provided and bos_text_provided else (
                'eos_text' if eos_text_provided else 'bos_text'
            )
            warnings.warn(
                f'The provided tokenizer adds special tokens, but you also specified {message}. This may result '
                +
                'in duplicated special tokens. Please be sure this is what you intend.',
            )

    @abstractmethod
    def __iter__(self) -> Iterable[Union[dict[str, bytes], dict[str, NDArray]]]:
        pass


class ConcatTokensDataset(AbstractConcatTokensDataset):
    """An IterableDataset that returns token samples for MDSWriter.

    Returns dicts of {'tokens': ndarray:int32}

    To use data created by this class and written to MDS format:

    ```python
        import torch
        from streaming.base import StreamingDataset
        from transformers import AutoTokenizer

        tokenizer = AutoTokenizer.from_pretrained('your/tokenizer')
        ds = StreamingDataset(local='mds-data-folder', split='val')

        # note, you need to copy the numpy array because the original is non-writeable
        # and torch does not support non-writeable tensors, so you get a scary warning and
        # if you do try to write to the tensor you get undefined behavior
        tokens = torch.from_numpy(np.frombuffer(ds[0]['tokens'], dtype=np.int32).copy())
        print(tokenizer.decode(tokens))
    ```
    """

    def __init__(
        self,
        hf_dataset: Union[hf_datasets.IterableDataset, hf_datasets.Dataset],
        tokenizer: PreTrainedTokenizerBase,
        max_length: int,
        bos_text: str,
        eos_text: str,
        no_wrap: bool,
    ):
        self.hf_dataset = hf_dataset
        super().__init__(tokenizer, max_length, bos_text, eos_text, no_wrap)

    def __iter__(self) -> Iterable[dict[str, NDArray]]:
        buffer = []
        for sample in self.hf_dataset:
            encoded = self.tokenizer(
                sample['text'],
                truncation=False,
                padding=False,
            )
            iids = encoded['input_ids']
            buffer = buffer + self.bos_tokens + iids + self.eos_tokens
            while len(buffer) >= self.max_length:
                concat_sample = buffer[:self.max_length]
                buffer = buffer[self.max_length:] if self.should_wrap else []
                yield {
                    # convert to ndarray to store in MDS format
                    'tokens': np.asarray(concat_sample, dtype=np.int32),
                }


def stream_remote_local_validate(
    remote: Optional[str],
    local: Optional[str],
    split: Optional[str],
):
    """Check that, if needed, the local/split directory exists.

    Args:
        remote (Optional[str]): Remote path to the dataset.
        local (Optional[str]): Local path to the dataset.
        split (Optional[str]): Subdirectory specifying which dataset split to use, if any.
    """
    if remote is None or (local == remote):
        if local is not None and os.path.isdir(local):
            contents = set(os.listdir(local))
            if split is not None and split not in contents:
                raise ValueError(
                    f'Local directory {local} does not contain split {split}',
                )



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import os
from typing import Optional

PROMPTFILE_PREFIX = 'file::'

__all__ = [
    'load_prompts',
    'load_prompts_from_file',
]


def load_prompts(prompts: list[str],
                 prompt_delimiter: Optional[str] = None) -> list[str]:
    """Loads a set of prompts, both free text and from file.

    Args:
        prompts (List[str]): List of free text prompts and prompt files
        prompt_delimiter (Optional str): Delimiter for text file
            If not provided, assumes the prompt file is a single prompt (non-delimited)

    Returns:
        List of prompt string(s)
    """
    prompt_strings = []
    for prompt in prompts:
        if prompt.startswith(PROMPTFILE_PREFIX):
            prompts = load_prompts_from_file(prompt, prompt_delimiter)
            prompt_strings.extend(prompts)
        else:
            prompt_strings.append(prompt)
    return prompt_strings


def load_prompts_from_file(
    prompt_path: str,
    prompt_delimiter: Optional[str] = None,
) -> list[str]:
    """Load a set of prompts from a text fie.

    Args:
        prompt_path (str): Path for text file
        prompt_delimiter (Optional str): Delimiter for text file
            If not provided, assumes the prompt file is a single prompt (non-delimited)

    Returns:
        List of prompt string(s)
    """
    if not prompt_path.startswith(PROMPTFILE_PREFIX):
        raise ValueError(f'prompt_path_str must start with {PROMPTFILE_PREFIX}')

    _, prompt_file_path = prompt_path.split(PROMPTFILE_PREFIX, maxsplit=1)
    prompt_file_path = os.path.expanduser(prompt_file_path)
    if not os.path.isfile(prompt_file_path):
        raise FileNotFoundError(
            f'{prompt_file_path=} does not match any existing files.',
        )

    with open(prompt_file_path, 'r') as f:
        prompt_string = f.read()

    if prompt_delimiter is None:
        return [prompt_string]
    return [i for i in prompt_string.split(prompt_delimiter) if i]



# Copyright 2022 MosaicML LLM Foundry authors
# SPDX-License-Identifier: Apache-2.0

import logging
import os

from composer.loggers import MosaicMLLogger
from composer.loggers.mosaicml_logger import (
    MOSAICML_ACCESS_TOKEN_ENV_VAR,
    MOSAICML_PLATFORM_ENV_VAR,
)

__all__ = [
    'SpecificWarningFilter',
]


class SpecificWarningFilter(logging.Filter):

    def __init__(self, message_to_suppress: str):
        """Filter out a specific warning message based on its content.

        This can be useful for filtering out specific warning messages from third party packages.

        Args:
            message_to_suppress (str): The warning message to suppress.
        """
        super().__init__()
        self.message_to_suppress = message_to_suppress

    def filter(self, record: logging.LogRecord) -> bool:
        return self.message_to_suppress not in record.getMessage()


def get_mosaicml_logger():
    if os.environ.get(MOSAICML_PLATFORM_ENV_VAR, 'false').lower(
    ) == 'true' and os.environ.get(MOSAICML_ACCESS_TOKEN_ENV_VAR):
        # Adds mosaicml logger to composer if the run was sent from Mosaic platform, access token is set
        return MosaicMLLogger()
    else:
        return None
